//
// "This program may be used, executed, copied, modified and distributed without royalty for the 
// purpose of developing, using, marketing, or distributing."
//
// Product 5724i63,  (C) COPYRIGHT International Business Machines Corp., 2001,2004
// All Rights Reserved * Licensed Materials - Property of IBM
// US Government Users Restricted Rights - Use, duplication or disclosure
// restricted by GSA ADP Schedule Contract with IBM Corp.
//
package com.ibm.websphere.samples.asynchbeans.wstrader;

import java.util.StringTokenizer;

/**
 * A simple response message interface.  Enforces a common message
 * interface to the remote clients.
 */
public class ClientInstruction
{

    public final static String RESPONSE_ERRORTYPE_ERRORMESSAGE = "errormsg";
    public final static String RESPONSE_ERRORTYPE_EXCEPTION = "exception";

    public final static String RESPONSE_ERRORMSG_NOLOGIN = "Missing login information";
    public final static String RESPONSE_ERRORMSG_NOSTREAMER = "No active streamer.";

    public final static String RESPONSE_TYPE_TICKER = "ticker";
    public final static String RESPONSE_TYPE_STATUS = "status";
    public final static String RESPONSE_TYPE_ERROR = "error";
    public final static String RESPONSE_TYPE_COMMANDSTATUS= "cmdstatus";
    
    public final static String RESPONSE_DELIMITER = "|";

    public final static String RESPONSE_COMMAND_LOGIN = "login";
    public final static String RESPONSE_COMMAND_LOGOUT = "logout";
    public final static String RESPONSE_COMMAND_UPDATE = "update";
    public final static String RESPONSE_COMMANDSTATUS_SUCCESS = "success";
    public final static String RESPONSE_COMMANDSTATUS_FAIL = "fail";
    
    public final static String RESPONSE_STATUS_VALUE_ALIVE = "alive";
    public final static String RESPONSE_STATUS_VALUE_FRESH = "fresh";
    public final static String RESPONSE_STATUS_VALUE_STALE = "stale";
    public final static String RESPONSE_STATUS_VALUE_DEAD = "dead";

    public final static String RESPONSE_STATUS_SOURCE_PRICEGEN = "PRICEGEN";
    public final static String RESPONSE_STATUS_SOURCE_STREAMSERVER = "STREAMSERVER";
    
    private String responseSeparator;

    /**
     * A response type can be [ping,ticker,error]
     */
    private String responseType;
    
    // Ticker information
    private String tickerSymbol;
    private String tickerType;
    private double tickerPrice;
    private String tickerTime;
    private double bidPrice;
    private double askPrice;
    
    // Status information
    private String statusSource;
    private String statusValue;
    
    // Error information
    private String errorType;
    private String errorMessage;
    private Throwable errorException;

    // Command status information
    private String command;
    private String commandStatus;
    private String commandUserData;

    private ClientInstruction(String responseType)
    {
        this.responseType = responseType;
        byte[] eol = {10};
        responseSeparator = new String(eol);
    }
    
    /**
     * Attempt to parse a line of text into a ClientInstruction object.
     * Return a null if the input string to parse is null or if 
     * the string is invalid.
     */
    public static ClientInstruction createResponseFromString(String resp)
    {
        if (resp == null || resp.length() <= 0)
        {
            return null;
        }
        
        StringTokenizer st = new StringTokenizer(resp, RESPONSE_DELIMITER);
        
        ClientInstruction r = new ClientInstruction(st.nextToken());
        
        if (r.responseType.equals(RESPONSE_TYPE_TICKER))
        {
            r.tickerSymbol = st.nextToken();
            r.tickerType = st.nextToken();
            r.tickerPrice = Double.valueOf(st.nextToken()).doubleValue();
            r.bidPrice = Double.valueOf(st.nextToken()).doubleValue();
            r.askPrice = Double.valueOf(st.nextToken()).doubleValue();
            r.tickerTime = st.nextToken();
        }
        else if (r.responseType.equals(RESPONSE_TYPE_ERROR))
        {
            r.errorType = st.nextToken();
            r.errorMessage = st.nextToken();
        }        
        else if (r.responseType.equals(RESPONSE_TYPE_STATUS))
        {
            r.statusSource = st.nextToken();
            r.statusValue = st.nextToken();
        }        
        else if (r.responseType.equals(RESPONSE_TYPE_COMMANDSTATUS))
        {
            r.command = st.nextToken();
            r.commandStatus = st.nextToken();
    
            if(st.hasMoreTokens())
            {
                r.commandUserData = st.nextToken();
            }
            else
            {
                r.commandUserData = null;
            }
        }        
        else
        {
            return null;
        }        
        
        return r;
        
    }
    
    /**
     * Create a Ticker price update response
     * @param tickerSymbol
     * @param tickerType
     * @param tickerPrice
     * @param bidPrice
     * @param askPrice
     * @param tickerTime
     * @return ClientInstruction
     */
    public static ClientInstruction createTickerResponse(String tickerSymbol, String tickerType, double tickerPrice, double bidPrice, double askPrice, String tickerTime)
    {
        ClientInstruction r = new ClientInstruction(RESPONSE_TYPE_TICKER);
        r.tickerSymbol = tickerSymbol;
        r.tickerType = tickerType;
        r.tickerPrice = tickerPrice;   
        r.bidPrice = bidPrice;
        r.askPrice = askPrice;
        r.tickerTime = tickerTime;
        return r;
    }
    
    /**
     * Create an Error response 
     * @param errorMessage
     * @return ClientInstruction
     */
    public static ClientInstruction createErrorResponse(String errorMessage)
    {
        ClientInstruction r = new ClientInstruction(RESPONSE_TYPE_ERROR);
        r.errorType = RESPONSE_ERRORTYPE_ERRORMESSAGE;
        r.errorMessage = errorMessage;
        return r;
    }

    /**
     * Create an Error response for a given exception
     * @param errorMessage
     * @param e
     * @return ClientInstruction
     */
    public static ClientInstruction createErrorResponse(String errorMessage, Throwable e)
    {
        ClientInstruction r = new ClientInstruction(RESPONSE_TYPE_ERROR);
        r.errorType = RESPONSE_ERRORTYPE_EXCEPTION;
        r.errorMessage = errorMessage;
        r.errorException = e;
        return r;
    }

    /**
     * Create a Status response 
     * @return ClientInstruction
     */
    public static ClientInstruction createStatusResponse(String statusSource, String statusValue)
    {
        ClientInstruction r = new ClientInstruction(RESPONSE_TYPE_STATUS);
        r.statusSource = statusSource;
        r.statusValue = statusValue;
        return r;
    }

    /**
     * Create a response to a command
     * @param command
     * @param status
     * @return ClientInstruction
     */
    public static ClientInstruction createCommandStatusResponse(String command, String status)
    {
        ClientInstruction r = new ClientInstruction(RESPONSE_TYPE_COMMANDSTATUS);
        r.command = command;
        r.commandStatus = status;
        r.commandUserData = null;
        return r;
    }    

    /**
     * Create a response to a command with user data
     * @param command
     * @param status
     * @param userData
     * @return ClientInstruction
     */
    public static ClientInstruction createCommandStatusResponse(String command, String status, String userData)
    {
        ClientInstruction r = new ClientInstruction(RESPONSE_TYPE_COMMANDSTATUS);
        r.command = command;
        r.commandStatus = status;
        r.commandUserData = userData;
        return r;
    }    

    /**
     * Assemble a simple response that is delimited by an EOL character.
     * @return String
     */
    public String getTextResponse()
    {
        StringBuffer resp = new StringBuffer(responseType);
        
        if (responseType.equals(RESPONSE_TYPE_TICKER))
        {
	        resp.append(RESPONSE_DELIMITER);
	        resp.append(tickerSymbol);
            resp.append(RESPONSE_DELIMITER);
            resp.append(tickerType);
            resp.append(RESPONSE_DELIMITER);
            resp.append(tickerPrice);
            resp.append(RESPONSE_DELIMITER);
            resp.append(bidPrice);
            resp.append(RESPONSE_DELIMITER);
            resp.append(askPrice);
            resp.append(RESPONSE_DELIMITER);
            resp.append(tickerTime);
        }
        else if (responseType.equals(RESPONSE_TYPE_ERROR))
        {
            resp.append(RESPONSE_DELIMITER);
            resp.append(errorType);
	        resp.append(RESPONSE_DELIMITER);
	        resp.append(errorMessage);
            if (errorType.equals(RESPONSE_ERRORTYPE_EXCEPTION))
            {
                resp.append(RESPONSE_DELIMITER);
                resp.append(errorException.toString());
            }
        }
        else if (responseType.equals(RESPONSE_TYPE_STATUS))
        {
            resp.append(RESPONSE_DELIMITER);
            resp.append(statusSource);
            resp.append(RESPONSE_DELIMITER);
            resp.append(statusValue);
        }
        else if (responseType.equals(RESPONSE_TYPE_COMMANDSTATUS))
        {
            resp.append(RESPONSE_DELIMITER);
            resp.append(command);
            resp.append(RESPONSE_DELIMITER);
            resp.append(commandStatus);
            if (commandUserData != null)
            {
	            resp.append(RESPONSE_DELIMITER);
	            resp.append(commandUserData);
            }
        }
        
        resp.append(responseSeparator);
        
        return resp.toString();
    }

	/**
	 * @see Object#toString()
	 */
	public String toString() {
		return getTextResponse();
	}

    public String getTickerSymbol()
    {
        return tickerSymbol;
    }

    public String getTickerType()
    {
        return tickerType;
    }

    public double getTickerPrice()
    {
        return tickerPrice;
    }
    
    public double getAskPrice()
    {
        return askPrice;
    }
    
    public double getBidPrice()
    {
        return bidPrice;
    }

    public String getTickerTime()
    {
        return tickerTime;
    }
        
    public String getErrorType()
    {
        return errorType;
    }

    public String getErrorMessage()
    {
        return errorMessage;
    }

    public Throwable getErrorException()
    {
        return errorException;
    }

    public String getCommand()
    {
        return command;
    }

    public String getCommandStatus()
    {
        return commandStatus;
    }
    
    public String getCommandUserData()
    {
        return commandUserData;
    }
    
    public String getResponseType()
    {
        return responseType;
    }

    public String getStatusSource()
    {
        return statusSource;
    }
    
    public String getStatusValue()
    {
        return statusValue;
    }
    
}
