//
// "This program may be used, executed, copied, modified and distributed without royalty for the 
// purpose of developing, using, marketing, or distributing."
//
// Product 5724i63,  (C) COPYRIGHT International Business Machines Corp., 2001,2004
// All Rights Reserved * Licensed Materials - Property of IBM
// US Government Users Restricted Rights - Use, duplication or disclosure
// restricted by GSA ADP Schedule Contract with IBM Corp.
//
package com.ibm.websphere.samples.asynchbeans.wstrader.applet;

import java.util.Enumeration;
import java.util.Vector;

import com.ibm.websphere.samples.asynchbeans.wstrader.ClientInstruction;

/**
 * A ServerMonitor is a singleton object (one per JVM) that monitors the status 
 * of a server.  It is very similar to the Asynch Beans SubsystemMonitor and Alarm
 * interfaces.  It publishes events based on the frequency of heartbeats that are
 * received from the ping() method.  
 * 
 * The monitor is started when a listener is registered with the addNotificationListener()
 * method and is destroyed when the last listener is removed with the removeNotificationListener() 
 * method.
 */
public class ServerMonitor extends Thread implements NotificationBroadcaster
{
    int interval;
    int staleBeats;
    int deadBeats;
    
    int beatsUntilStale;
    int beatsUntilDead;

    int status;
    final int ALIVE = 0;
    final int STALE = 1;
    final int DEAD = 2;

    /**
     * If true, the thread has been asked to end.
     */
    private boolean isCancelled;
    
    /**
     * If true, send debug messages to the Java Console
     */
    private boolean isDebug;

    /**
     * The listeners who have subscribed to notification events from this object.
     */
    private static Vector listeners = new Vector();
    
    /**
     * The singleton ServerMonitor.
     */
    private static ServerMonitor self = null;

    /**
     * Constuct a new ServerMonitor
     * @param interval_ms the number of milliseconds between expected ping() calls.  The heartbeat interval.
     * @param staleBeats the number of heartbeats that can be missed before calling the Stale event.
     * @param deadBeats the number of heartbeats that can be missed before calling the Dead event.
     */
    private ServerMonitor(int interval_ms, int staleBeats, int deadBeats)
    {
        this.interval = interval_ms;
        this.staleBeats = staleBeats;
        this.deadBeats = deadBeats;
        beatsUntilStale = staleBeats;
        beatsUntilDead = deadBeats;
        status = DEAD;
        isCancelled = false;
        isDebug = false;
    }
    
    /**
     * Create a new singleton ServerMonitor or get the currently created on if one already exists
     */
    public static synchronized ServerMonitor create(int interval_ms, int staleBeats, int deadBeats)
    {
        if (self == null)
        {
            self = new ServerMonitor(interval_ms, staleBeats, deadBeats);
        }
        
        return self;
    }
    
    /**
     * Get the singleton ServerMonitor
     */
    public static synchronized ServerMonitor getSelf()
    {
        return self;
    }
    
    /**
     * Reset the state of the monitor to Alive and fire a fresh event if the current state
     * is stale or dead.
     */
    public void ping()
    {
        boolean doFireFresh = false;
        synchronized(this)
        {
            beatsUntilStale = staleBeats;
            beatsUntilDead = deadBeats;
            if(status == STALE || status == DEAD)
            {
                doFireFresh = true;
            }
            status = ALIVE;
        }
    
        if (doFireFresh)
        {
            fireFresh();
        }
    }

    /**
     * Each x milliseconds (heartbeat interval), decrement the number of beats it will 
     * take to become stale or dead.  Once this threshold is reached, change
     * our status and fire the event.
     */
    public void run()
    {
        try
        {
            this.setName("ServerMonitor");
            while(!getCancelled() && !interrupted())
            {
                try
                {
                    sleep(interval);
                    
                    boolean doFireDead = false;
                    boolean doFireStale = false;
                    
                    synchronized(this)
                    {
                        --beatsUntilStale;
                        --beatsUntilDead;
                        
                        if (isDebug)
                        {
                            System.out.println("ServerMonitor.run(): " + status + ", " + beatsUntilStale + ", " + beatsUntilDead);
                        }
                        if(status == STALE && beatsUntilDead < 0)
                        {
                            status = DEAD;
                            doFireDead = true;
                        }
                        else if (status == ALIVE && beatsUntilStale < 0)
                        {
                            status = STALE;
                            doFireStale = true;
                        }
    
                    }                    
                    if (doFireDead)
                        fireDead();
                    else if (doFireStale)
                        fireStale();
                }
                catch(InterruptedException e)
                {
                    System.out.println("ServerMonitor was interrupted.");
                }
            }
        }
        finally
        {
        }
    }
    
    private void fireFresh()
    {
        fireNotification(ClientInstruction.createStatusResponse(ClientInstruction.RESPONSE_STATUS_SOURCE_STREAMSERVER, ClientInstruction.RESPONSE_STATUS_VALUE_FRESH));
    }
    
    private void fireStale()
    {
        fireNotification(ClientInstruction.createStatusResponse(ClientInstruction.RESPONSE_STATUS_SOURCE_STREAMSERVER, ClientInstruction.RESPONSE_STATUS_VALUE_STALE));
    }
    
    private void fireDead()
    {
        fireNotification(ClientInstruction.createStatusResponse(ClientInstruction.RESPONSE_STATUS_SOURCE_STREAMSERVER, ClientInstruction.RESPONSE_STATUS_VALUE_DEAD));
    }

    /**
     * @see com.ibm.websphere.samples.asynchbeans.wstrader.applet.NotificationBroadcaster#addNotificationListener(NotificationListener)
     */
    public synchronized void addNotificationListener(NotificationListener listener)
    {
        listeners.addElement(listener);

        // If this thread isn't started, do so.
        if (listeners.size() == 1)
        {
            start();
        }
    }

    /**
     * @see com.ibm.websphere.samples.asynchbeans.wstrader.applet.NotificationBroadcaster#removeNotificationListener(NotificationListener)
     */
    public synchronized void removeNotificationListener(NotificationListener listener)
    {
        listeners.removeElement(listener);
        
        // If there are no more listeners, cancel ourselves.
        if(listeners.size() == 0)
        {
            _destroy();
        }
    }

    /**
     * Fire a notification to all of our subscribed listeners.
     * @param ci the ClientInstruction that our subscribers can consume.  We are just reusing the class that is
     *         normally used for sending instructions from a Servlet to a Client.
     */
    private void fireNotification(ClientInstruction ci)
    {        
        System.out.println("ServerMonitor.fireNotification(): " + ci.getStatusValue());
        // Notify all listeners of the result of the LOGOUT
        for (Enumeration e = listeners.elements();e.hasMoreElements();)
        {
            NotificationListener l = (NotificationListener) e.nextElement();
            Notification n = new Notification(Notification.NOTIFICATION_STATUS, this);
            n.setUserData(ci);
            l.handleNotification(n);
        }    
    }

    /**
     * Fire an event with our current status.
     */
    public void fireCurState()
    {
        int curStatus = getStatus();
        if (curStatus == ALIVE)
        {
            fireFresh();
        }
        else if (curStatus == STALE)
        {
            fireStale();
        }
        else if (curStatus == DEAD)
        {
            fireDead();
        }
    }

    /**
     * Cancel this monitor.
     */     
    public synchronized void cancel()
    {
        isCancelled = true;
    }

    /**
     * Find out if the thread has been requested to end.
     */    
    public synchronized boolean getCancelled()
    {
        return isCancelled;
    }


    /**
     * Returns the status.
     * @return int
     */
    public synchronized int getStatus()
    {
        return status;
    }

    /**
     * Sets the status.
     * @param status The status to set
     */
    public synchronized void setStatus(int status)
    {
        this.status = status;
    }

    /**
     * Returns the isDebug.
     * @return boolean
     */
    public boolean isDebug()
    {
        return isDebug;
    }

    /**
     * Sets the isDebug.
     * @param isDebug The isDebug to set
     */
    public void setIsDebug(boolean isDebug)
    {
        this.isDebug = isDebug;
    }

    /**
     * Destroy this instance of the ServerMonitor.
     */
    private synchronized void _destroy()
    {
        // Cancel our thread and then reset our static variables
        cancel();
        self = null;
    }


}
