//
// "This program may be used, executed, copied, modified and distributed without royalty for the 
// purpose of developing, using, marketing, or distributing."
//
// Product 5724i63,  (C) COPYRIGHT International Business Machines Corp., 2001,2004
// All Rights Reserved * Licensed Materials - Property of IBM
// US Government Users Restricted Rights - Use, duplication or disclosure
// restricted by GSA ADP Schedule Contract with IBM Corp.
//
package com.ibm.websphere.samples.asynchbeans.wstrader.applet;

import java.io.EOFException;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLConnection;
import java.util.Enumeration;
import java.util.Vector;

import com.ibm.websphere.samples.asynchbeans.wstrader.ClientInstruction;
import com.ibm.websphere.samples.asynchbeans.wstrader.Config;

/**
 * Sends information to our various Ticker Streamer servlets by queuing
 * all requests to run on a single thread.  We use a singleton, so 
 * all applets running in the same JVM are using a single HTTP connection.
 */
public class URLSenderHandler extends Thread
{
    /**
     * A queue of messages.  We can only send one message at a time.
     */
    private static Vector messageQueue = new Vector();
    
    /**
     * The code base of our applet.  Used for determining the servlet to 
     * connect to.
     */
    private static URL codeBase = null;
    
    /**
     * A singleton URLSenderHandler reference
     */
    private static URLSenderHandler self = null;
    
    /**
     * The listeners who have subscribed to notification events from this object.
     */
    private static Vector listeners = new Vector();
    
    /**
     * A basic utility class instance.
     */
    private static Utility utility = new Utility();
    
    /**
     * If true, the thread should terminate itself.
     */
    private boolean isCancelled;    
    
    /**
     * Create a URLSenderHandler
     * @param codeBase the calling applet's codeBase.
     */
    private URLSenderHandler(URL codeBase)
    {
        this.codeBase = codeBase;
    }
    
    /**
     * Return a singleton instance of the URLSenderHandler.
     * @return URLSenderHandler object.
     */
    public static synchronized URLSenderHandler getSelf()
    {
        return self;
    }
    
    /**
     * Initialize the singleton URLSenderHandler.
     * @param codeBase the calling applet's codeBase.
     */
    public static synchronized URLSenderHandler create(URL codeBase)
    {
        if (self == null)
        {
            self = new URLSenderHandler(codeBase);
        }
        return self;
    }
    
    /**
     * Add a message to the queue of messages and notify the singleton monitor.
     * @param messageItem the message to queue.
     */
    private static void enQueue(MessageItem messageItem)
    {
        synchronized(self)
        {
	        messageQueue.addElement(messageItem);
	        self.notify();
        }
    }
    
    /**
     * Remove a message from the queue of messages.  If the queue is empty, then 
     * block until a new message is added
     */
    private static MessageItem deQueue()
        throws InterruptedException
    {
        synchronized(self)
        {
            // block if the message queue is empty
            if(messageQueue.size() == 0)
            {
                self.wait();
            }

            // If not empty, then grab the message off the queue
            MessageItem msg = (MessageItem) messageQueue.elementAt(0);
            messageQueue.removeElementAt(0);
            return msg;
        }
    }
    
    /**
     * Send a Ping message to the TickerStreamer server app.
     * @param cookie the Cookie receieved from the login.
     */
    public static void sendPingMessage(String cookie)
    {
        MessageItem msg = MessageItem.getPingMessage(cookie);
        enQueue(msg);
    }
    
    /**
     * Send a Client Login message to the TickerStreamer server app.
     * @param clientid the client id to login as
     * @param symbols the Ticker Symbols to use if this client is not already
     *         logged-in.
     */
    public static void sendLoginMessage(String clientid, String symbols)
    {
        MessageItem msg = MessageItem.getLoginMessage(clientid, symbols);
        enQueue(msg);
    }
    
    /**
     * Send a Client Logout message to the TickerStreamer server app.
     * @param cookie the Cookie receieved from the login.
     */
    public static void sendLogoutMessage(String cookie)
    {
        MessageItem msg = MessageItem.getLogoutMessage(cookie);
        enQueue(msg);
    }
    
    /**
     * Send a Change Tickers message to the TickerStreamer server app.
     * @param streamerName the streamer to update (example: "DEFAULT")
     * @param cookie the Cookie receieved from the login.
     * @param symbols the Ticker Symbols to update to.
     */
    public static void sendUpdateMessage(String streamerName, String cookie, String symbols)
    {
        MessageItem msg = MessageItem.getUpdateMessage(streamerName, cookie, symbols);
        enQueue(msg);
    }
    
    /**
     * Create a URL object for this site using the given servletname
     * @param servletName the name of the servlet to talk to.
     */
    URL constructURL(String servletName)
        throws MalformedURLException
    {
        URL url = new URL(codeBase.getProtocol(), codeBase.getHost(), codeBase.getPort(), servletName);
        return url;
    }
    
    /**
     * Ping the WebSphere Trader application using the Ping servlet
     * @param cookie the Cookie receieved from the login.
     */
    private void pingServer(String cookie)
        throws IOException
    {
        // Ping
        URL pingURL = constructURL("/asynchbeans/WebSphereTrader/ping");
        System.out.println("Pinging server: " + pingURL);
        URLConnection conn = (URLConnection) pingURL.openConnection();

        conn.setUseCaches(false);
        conn.setAllowUserInteraction(false);
        conn.setRequestProperty("Cookie", cookie);
        
        conn.connect();
        InputStream is = conn.getInputStream();
        InputStreamReader isr = new InputStreamReader(is);

        Object content = conn.getContent();
        
        // Close the connections
        is.close();
        System.out.println("Ping complete.");
    }

    /**
     * Login to the WebSphere Trader application by way of the login servlet.
     * @param clientid the client id to login as
     * @param symbols the Ticker Symbols to use if this client is not already
     *         logged-in.
     */
    private void doLogin(String clientid, String symbols)
    {
        // Login.
        InputStream is = null;
        boolean rc = false;
        String cookie = null;


        try
        {
            URL loginURL = constructURL("/asynchbeans/WebSphereTrader/login?" + Config.CLIENT_PARAM_CLIENTID + "=" + java.net.URLEncoder.encode(clientid));
            URLConnection conLogin = (URLConnection) loginURL.openConnection();
    
            conLogin.setUseCaches(false);
            conLogin.setAllowUserInteraction(false);
            
            is = conLogin.getInputStream();
            InputStreamReader isr = new InputStreamReader(is);
            
            Object content = conLogin.getContent();
        
            
            // Process the data in our login response
            try
            {
                String line = utility.getLineFromInputStream(isr);        
        
                ClientInstruction resp = ClientInstruction.createResponseFromString(line);
                if (resp != null)
                {
                    String output = resp.getTextResponse();
                    System.out.println(output);
                    if (resp.getResponseType().equals(ClientInstruction.RESPONSE_TYPE_COMMANDSTATUS))
                    {
                        if(resp.getCommand().equals(ClientInstruction.RESPONSE_COMMAND_LOGIN))
                        {
                            if(resp.getCommandStatus().equals(ClientInstruction.RESPONSE_COMMANDSTATUS_SUCCESS))
                            {
                                // If the client WAS already active, then use the list of symbols received from the server
                                if (resp.getCommandUserData() != null)
                                {
                                    symbols = resp.getCommandUserData();
                                }
                                rc = true;
                            }
                        }
                    }
                }
    
    
                cookie = conLogin.getHeaderField("Set-Cookie");
    
            }
            catch(EOFException ex)
            {
                // Just absorb this 
            }
        }
        catch(IOException ex)
        {
            ex.printStackTrace();
            rc = false;
        }
        
        // Notify all listeners of the result of the login
        for (Enumeration e = listeners.elements();e.hasMoreElements();)
        {
            NotificationListener l = (NotificationListener) e.nextElement();
            Notification n = new Notification(Notification.NOTIFICATION_LOGIN, this);
            LoginInfo loginInfo = new LoginInfo();
            loginInfo.cookie = cookie;
            loginInfo.symbols = symbols;
            loginInfo.success = rc;
            n.setUserData(loginInfo);
            l.handleNotification(n);
        }
    }

    /**
     * Logout from the WebSphere Trader application by way of the logout servlet.
     * @param cookie the Cookie receieved from the login.
     */
    public boolean doLogout(String cookie)
        throws MalformedURLException, IOException
    {
        boolean rc = false;    

        if (cookie != null)
        {
            URL updateURL = constructURL("/asynchbeans/WebSphereTrader/logout");
            System.out.println("Opening logout url: " + updateURL.toExternalForm());
            URLConnection conLogout = (URLConnection) updateURL.openConnection();
    
            conLogout.setUseCaches(false);
            conLogout.setRequestProperty("Cookie", cookie);
            conLogout.setAllowUserInteraction(false);
            
            InputStream is = conLogout.getInputStream();
            InputStreamReader isr = new InputStreamReader(is);
            
            Object content = conLogout.getContent();
        
            // Process the data in our logout response
            try
            {
                String line = utility.getLineFromInputStream(isr);        
        
                ClientInstruction resp = ClientInstruction.createResponseFromString(line);
                if (resp != null)
                {
                    String output = resp.getTextResponse();
                    System.out.println(output);
                    if (resp.getResponseType().equals(ClientInstruction.RESPONSE_TYPE_COMMANDSTATUS))
                    {
                        if(resp.getCommand().equals(ClientInstruction.RESPONSE_COMMAND_LOGOUT))
                        {
                            if(resp.getCommandStatus().equals(ClientInstruction.RESPONSE_COMMANDSTATUS_SUCCESS))
                            {
                                rc = true;
                            }
                        }
                    }
                }
                
            }
            catch(EOFException ex)
            {
                // Just absorb this 
            }
            
    
            // Close the connections
            is.close();
            System.out.println("Logout Complete");
        }

        // Notify all listeners of the result of the LOGOUT
        for (Enumeration e = listeners.elements();e.hasMoreElements();)
        {
            NotificationListener l = (NotificationListener) e.nextElement();
            Notification n = new Notification(Notification.NOTIFICATION_LOGOUT, this);
            n.setUserData(new Boolean(rc));
            l.handleNotification(n);
        }

        return rc;
    }

    /**
     * Update the current client's ticker symbols on the WebSphere Trader application 
     * by way of the update servlet.
     * @param streamerName the streamer to update (example: "DEFAULT")
     * @param cookie the Cookie receieved from the login.
     * @param symbols the Ticker Symbols to update to.
     */
    public boolean doUpdate(String streamerName, String cookie, String symbols)
        throws MalformedURLException, IOException
    {
        boolean rc = false;    

        if (cookie != null)
        {
            URL updateURL = constructURL("/asynchbeans/WebSphereTrader/update?" + Config.CLIENT_PARAM_STREAM_NAME + "=" + java.net.URLEncoder.encode(streamerName) + "&TSTickers="+ java.net.URLEncoder.encode(symbols));
            System.out.println("Opening update url: " + updateURL.toExternalForm());
            URLConnection conUpdate = (URLConnection) updateURL.openConnection();
    
            conUpdate.setUseCaches(false);
            conUpdate.setRequestProperty("Cookie", cookie);
            conUpdate.setAllowUserInteraction(false);
            
            InputStream is = conUpdate.getInputStream();
            InputStreamReader isr = new InputStreamReader(is);
            
            Object content = conUpdate.getContent();
        
            // Process the data in our login response
            try
            {
                String line = utility.getLineFromInputStream(isr);        
        
                ClientInstruction resp = ClientInstruction.createResponseFromString(line);
                if (resp != null)
                {
                    String output = resp.getTextResponse();
                    System.out.println(output);
                    if (resp.getResponseType().equals(ClientInstruction.RESPONSE_TYPE_COMMANDSTATUS))
                    {
                        if(resp.getCommand().equals(ClientInstruction.RESPONSE_COMMAND_UPDATE))
                        {
                            if(resp.getCommandStatus().equals(ClientInstruction.RESPONSE_COMMANDSTATUS_SUCCESS))
                            {
                                rc = true;
                            }
                        }
                    }
                }
                
            }
            catch(EOFException ex)
            {
                // Just absorb this 
            }
            
    
            // Close the connections
            is.close();
            System.out.println("Update Complete");
        }

        // Notify all listeners of the result of the login
        for (Enumeration e = listeners.elements();e.hasMoreElements();)
        {
            NotificationListener l = (NotificationListener) e.nextElement();
            Notification n = new Notification(Notification.NOTIFICATION_UPDATE, this);
            n.setUserData(new Boolean(rc));
            l.handleNotification(n);
        }

        return rc;
    }

	/**
	 * Execute the URLSender loop.  Continually process messages from
     * our message queue such that only one message is processed at a time.
     * We do this to avoid starting more than one connection.  We will only have a single
     * HTTP connection open at a given time.
	 */
	public void run() {
	        
	    try
	    {
            setName("URLSenderHandler");
   	        while(!getCancelled() && !isInterrupted())
	        {
                // Receive a message from our queue.  This operation will
                // block until a message is available.
	            MessageItem msg = deQueue();
	
	            try
	            {
                    // Process the message type by calling the appropriate action method
		            if (msg.messageType == MessageItem.MTYPE_PING)
		            {
		                pingServer(msg.pingCookie);
		            }
                    else if (msg.messageType == MessageItem.MTYPE_LOGIN)
                    {
                        doLogin(msg.loginClientID, msg.loginSymbols);
                    }
                    else if (msg.messageType == MessageItem.MTYPE_UPDATE)
                    {
                        doUpdate(msg.updateStreamerName, msg.updateCookie, msg.updateSymbols);
                    }
                    else if (msg.messageType == MessageItem.MTYPE_LOGOUT)
                    {
                        doLogout(msg.logoutCookie);
                    }
                    
	            }
	            catch(IOException ex)
	            {
	                ex.printStackTrace();
	            }
	        }
		}
        catch(InterruptedException ex)
        {
            System.out.println("URLSenderHander Interrupted");
        }
	    catch(Exception ex)
	    {
	        ex.printStackTrace();
	    }
    }



    /**
     * Cancel this monitor.
     */     
    public synchronized void cancel()
    {
        isCancelled = true;
    }

    /**
     * Find out if the thread has been requested to end.
     */    
    public synchronized boolean getCancelled()
    {
        return isCancelled;
    }
    
    /**
     * Add a listener for Notifications that will be recieved asynchronously.
     * @param listener the listener to notify.
     */
    public synchronized void addNotificationListener(NotificationListener listener)
    {
        listeners.addElement(listener);

        // If this thread isn't started, do so.
        if (listeners.size() == 1)
        {
            start();
        }
    }
    
    /**
     * Remove a listener from the Notification list.
     * @param listener 
     */
    public synchronized void removeNotificationListener(NotificationListener listener)
    {
        listeners.removeElement(listener);

        // If there are no more listeners, cancel ourselves.
        if(listeners.size() == 0)
        {
            _destroy();
        }
    }

    /**
     * Destroy this instance of the URLSenderHander.
     */
    private synchronized void _destroy()
    {
        // Cancel our thread and then reset our static variables
        cancel();
        self = null;
    }
}
