//
// "This program may be used, executed, copied, modified and distributed without royalty for the
// purpose of developing, using, marketing, or distributing."
//
// Product 5724i63,  (C) COPYRIGHT International Business Machines Corp., 2001,2004
// All Rights Reserved * Licensed Materials - Property of IBM
// US Government Users Restricted Rights - Use, duplication or disclosure
// restricted by GSA ADP Schedule Contract with IBM Corp.
//
package com.ibm.websphere.samples.asynchbeans.wstrader.pricegen;

import java.rmi.RemoteException;
import java.text.SimpleDateFormat;
import java.util.Date;

import javax.jms.*;
import javax.naming.InitialContext;
import javax.naming.NamingException;

import java.util.logging.*;
import com.ibm.websphere.asynchbeans.Alarm;
import com.ibm.websphere.asynchbeans.AlarmListener;

/**
 * An asynchronous Alarm that fires at a given interval and posts
 * a Ping message on a queue
 */
public class PriceHeartBeatAlarm implements AlarmListener
{
    static final Logger logger = Logger.getLogger(PriceHeartBeatAlarm.class.getName());
    static final String className = PriceHeartBeatAlarm.class.getName();

	String ticker;
	int interval;

    boolean isDebugEnabled = false;

	transient TopicSession jmsSession;
    transient TopicPublisher jmsPublisher;
    transient Topic topic;

    /**
     * Create a PricePublisherAlarm.  The alarm will fire, publishing ticker information to a JMS Topic.
     * The information and instructions on the ticker symbol, price information and how the price changes
     * is included in the data parameter. The topic connection is reestablished each time the alarm is fired instead
     * of storing it when this alarm is constructed.  This must be done because the topic connection would
     * be associated with the container that created the alarm and not with the alarm itself.
     * @param interval_ms the number of milliseconds to wait between ticks.
     * @param topic the JMS Topic to publish the ping messages on
     */
	public PriceHeartBeatAlarm(int interval_ms, Topic topic)
		throws RemoteException, JMSException
	{
        this.topic = topic;
		interval = interval_ms;
	}

	/**
	 * @see AlarmListener#fired(Alarm)
	 */
	public void fired(Alarm alarm)
	{
        String methodName = "fired";
        logger.entering(className, methodName);

        TopicConnection topicConnection = null;

		try
		{
            SimpleDateFormat tf = new SimpleDateFormat ("hh:mm:ss a");

            // Get jms items.
            InitialContext ic = new InitialContext();
            TopicConnectionFactory tcf = (TopicConnectionFactory)ic.lookup("java:comp/env/jms/TCF");
            topicConnection = tcf.createTopicConnection();
            topicConnection.start();

            jmsSession = topicConnection.createTopicSession(false, Session.AUTO_ACKNOWLEDGE);
            jmsPublisher = jmsSession.createPublisher(topic);
            jmsPublisher.setDeliveryMode(DeliveryMode.NON_PERSISTENT);
            jmsPublisher.setTimeToLive(2000);

			Message msg = jmsSession.createMessage();
			msg.setStringProperty("type", "ping");
            msg.setStringProperty("time", tf.format(new Date()));

            // Publish the message
            jmsPublisher.publish(msg);

	        // Reset the alarm for the designated interval, so it fires again.
	        alarm.reset(interval);
		}
		catch(java.lang.IllegalStateException e)
		{
			// if the AsynchScope owning the alarm was destroyed then
			// this exception is thrown when we call Alarm.reset so just
			// swallow it here so avoid any messages. It's expected.
		}
		catch(JMSException e)
		{
			logger.logp(Level.SEVERE, className, methodName, "JMSException during fired()", e);
		}
        catch(NamingException e)
        {
			logger.logp(Level.SEVERE, className, methodName, "JMSException during fired()", e);
        }
        finally
        {
            try
            {
                if (topicConnection != null)
                {
                    topicConnection.close();
                }
            }
            catch(JMSException e)
            {
               logger.logp(Level.SEVERE, className, methodName, "JMSException during fired()", e);
            }
            finally
            {
               logger.exiting(className, methodName);
            }
        }

	}
}
