//
// "This program may be used, executed, copied, modified and distributed without royalty for the
// purpose of developing, using, marketing, or distributing."
//
// Product 5724-J08, 5724-I63, 5724-H88, 5724-H89, 5655-N02, 5733-W70 COPYRIGHT International Business Machines Corp.  2001,2008
// All Rights Reserved * Licensed Materials - Property of IBM
// US Government Users Restricted Rights - Use, duplication or disclosure
// restricted by GSA ADP Schedule Contract with IBM Corp.
//
package com.ibm.websphere.samples.asynchbeans.wstrader.pricegen;

import java.rmi.RemoteException;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.Random;

import javax.jms.DeliveryMode;
import javax.jms.JMSException;
import javax.jms.Message;
import javax.jms.Topic;
import javax.jms.TopicConnection;
import javax.jms.TopicConnectionFactory;
import javax.jms.TopicPublisher;
import javax.jms.TopicSession;
import javax.naming.InitialContext;
import javax.naming.NamingException;

import java.util.logging.*;
import com.ibm.websphere.asynchbeans.Alarm;
import com.ibm.websphere.asynchbeans.AlarmListener;
import com.ibm.websphere.samples.asynchbeans.wstrader.Config;
import com.ibm.websphere.samples.asynchbeans.wstrader.pricegen.ejb.RawTickerConfigLocal;

/**
 * An asynchronous Alarm that fires at a given interval for a specific
 * ticker symbol.
 */
public class PricePublisherAlarm implements AlarmListener
{
    static final Logger logger = Logger.getLogger(PricePublisherAlarm.class.getName());
    static final String className = PricePublisherAlarm.class.getName();

	String ticker;
	double currentPrice;
	int decisionSpace;
	int downTickSpace;
	int upTickSpace;
	double tickPriceDelta;
    double lowestPrice;
    long tickPeriod;
	Random random;
	int interval;
	Topic topic;
    double bid;
    double ask;
    double spread;

	transient TopicSession jmsSession;
	transient TopicPublisher jmsPub;

    /**
     * Create a PricePublisherAlarm.  The alarm will fire, publishing ticker information to a JMS Topic
     * at a sepcified interval.  The topic connection is reestablished each time the alarm is fired instead
     * of storing it when this alarm is constructed.  This must be done because the topic connection would
     * be associated with the container that created the alarm and not with the alarm itself.
     * The information and instructions on the ticker symbol, price information and how the price changes
     * is included in the data parameter.
     * @param data the ticker configuration information that includes details on how to fire the ticker.
     * @param topic  the JMS Topic to post the ticker message to.
     */
	public PricePublisherAlarm(RawTickerConfigLocal data, Topic topic)
		throws RemoteException, JMSException
	{
		ticker = data.getTicker();
		currentPrice = data.getStartPrice().doubleValue();
		decisionSpace = data.getTickerDecisionRange().intValue();
		downTickSpace = data.getTickerDownTickRange().intValue();
		upTickSpace = data.getTickerUpTickRange().intValue();
		tickPriceDelta = data.getPriceDeltaByTick().doubleValue();
        tickPeriod = data.getTickPeriod().longValue();
        lowestPrice = currentPrice;
        spread = tickPriceDelta * 10;
        bid = currentPrice - spread / 2;
        ask = bid + spread;
        this.topic = topic;
		interval = (int)data.getTickPeriod().longValue();
		random = new Random();
	}

    /**
     * Setup the next bid and ask price and return a String indicating whether or not it
     * went up, down or stayed the same.
     */
    private String configureNextPrice()
    {
        String tickType = Config.PRICEGEN_PRICETYPE_STATIC;
        int n = random.nextInt(decisionSpace);
        if(n <= downTickSpace)
        {
            currentPrice -= tickPriceDelta;
            if(currentPrice <= 0)
            {
                currentPrice += tickPriceDelta*2;
            }
        }
        if(n >= upTickSpace)
        {
            currentPrice += tickPriceDelta;
        }

        if (lowestPrice > currentPrice)
        {
            lowestPrice = currentPrice;
        }

        // adjust the ask price if necessary
        if(currentPrice < (bid - tickPriceDelta*2))
        {
            bid = currentPrice;
            ask = bid + spread;
            tickType = Config.PRICEGEN_PRICETYPE_DOWN;
        }

        if(currentPrice > (ask + tickPriceDelta*2))
        {
            ask = currentPrice;
            bid = ask  - spread;
            tickType = Config.PRICEGEN_PRICETYPE_UP;
        }

        return tickType;
    }

	/**
	 * @see AlarmListener#fired(Alarm)
	 */
	public void fired(Alarm alarm)
	{
        TopicConnection topicConnection = null;

		try
		{
			// Adjust current price if necessary
			String tickType = configureNextPrice();

            SimpleDateFormat tf = new SimpleDateFormat ("hh:mm:ss a");


            InitialContext ic = new InitialContext();
            TopicConnectionFactory tcf = (TopicConnectionFactory)ic.lookup("java:comp/env/jms/TCF");
            topicConnection = tcf.createTopicConnection();
            topicConnection.start();

            jmsSession = topicConnection.createTopicSession(false, TopicSession.AUTO_ACKNOWLEDGE);
            jmsPub = jmsSession.createPublisher(topic);
            jmsPub.setDeliveryMode(DeliveryMode.NON_PERSISTENT);
            jmsPub.setTimeToLive(2000);

			Message msg = jmsSession.createMessage();
			msg.setStringProperty("ticker", ticker);
			msg.setDoubleProperty("price", currentPrice);
			msg.setStringProperty("type", tickType);
            msg.setDoubleProperty("bid", bid);
            msg.setDoubleProperty("ask", ask);
            msg.setStringProperty("time", tf.format(new Date()));

            logger.logp(Level.FINEST, className, "fired", "Publishing {0} {1} {2}: ", new Object[]{ticker, new Double(currentPrice), tickType});

            // Publish the message
			jmsPub.publish(msg);

	        // Reset the alarm for the designated interval, so it fires again.
	        alarm.reset(interval);
		}
		catch(java.lang.IllegalStateException e)
		{
			// if the AsynchScope owning the alarm was destroyed then
			// this exception is thrown when we call Alarm.reset so just
			// swallow it here so avoid any messages. It's expected.
		}
		catch(JMSException e)
		{
            logger.logp(Level.SEVERE, className, "fired", "JMSException during fired()", e);
		}
        catch(NamingException e)
        {
            logger.logp(Level.SEVERE, className, "fired", "NamingException during fired()", e);
        }
        finally
        {
            try
            {
                if (topicConnection != null)
                {
                    topicConnection.close();
                }
            }
            catch(JMSException e)
            {
                logger.logp(Level.SEVERE, className, "fired", "JMSException during fired()", e);
            }
        }


	}
}
