//
// "This program may be used, executed, copied, modified and distributed without royalty for the
// purpose of developing, using, marketing, or distributing."
//
// Product 5724i63,  (C) COPYRIGHT International Business Machines Corp., 2001,2004
// All Rights Reserved * Licensed Materials - Property of IBM
// US Government Users Restricted Rights - Use, duplication or disclosure
// restricted by GSA ADP Schedule Contract with IBM Corp.
//
package com.ibm.websphere.samples.asynchbeans.wstrader.pricegen.ejb;

import java.rmi.RemoteException;
import java.util.Collection;
import java.util.Iterator;

import javax.ejb.DuplicateKeyException;
import javax.ejb.EJBException;
import javax.ejb.SessionContext;
import javax.jms.Queue;
import javax.jms.Topic;
import javax.naming.InitialContext;
import javax.naming.NamingException;

import java.util.logging.*;
import com.ibm.websphere.asynchbeans.AlarmManager;
import com.ibm.websphere.asynchbeans.AsynchScope;
import com.ibm.websphere.asynchbeans.WorkManager;
import com.ibm.websphere.samples.asynchbeans.wstrader.Config;
import com.ibm.websphere.samples.asynchbeans.wstrader.pricegen.PriceGenConfig;
import com.ibm.websphere.samples.asynchbeans.wstrader.pricegen.PriceHeartBeatAlarm;
import com.ibm.websphere.samples.asynchbeans.wstrader.pricegen.PricePublisherAlarm;

/**
 * Contains all logic for starting and stopping a Price Generator.
 * For each price publisher configured in the RawTickerConfig entity bean,
 * we create alarms to publish each tick.
 */
public class PriceGenControlBean implements javax.ejb.SessionBean
{
    static final Logger logger = Logger.getLogger(PriceGenControlBean.class.getName());
    static final String className = PriceGenControlBean.class.getName();

    protected SessionContext mySessionCtx;

    /**
     * This method starts the price generator alarms when the application starts.
     * We get all the configured pricers in the database and create an Alarm for each
     * one to publish a message with the new synthetic price.
     * @return boolean True if the publishers were started, False if the publishers were already
     * started.
     */
    public boolean startPublishers() throws Exception
    {
        String methodName = "startPublishers";
        logger.entering(className, methodName);

        try
        {
            InitialContext ic = new InitialContext();
	      WorkManager wm = (WorkManager)ic.lookup("java:comp/env/wm/WorkManager");
            AsynchScope as = wm.findAsynchScope(PriceGenConfig.AS_NAME);

            if (as != null)
            {
                // Our bean already exists, so we must have already started.
                return false;
            }

        	as = wm.createAsynchScope(PriceGenConfig.AS_NAME);

            AlarmManager am = as.getAlarmManager();

            RawTickerConfigLocalHome rtcHome = (RawTickerConfigLocalHome)ic.lookup("java:comp/env/ejb/RawTickerConfig");

            Topic topic = (Topic)ic.lookup("java:comp/env/jms/PriceTopic");

            // Get the initial interval for the first publish alarm
            int interval = ((Integer)ic.lookup("java:comp/env/PricePublishInterval_ms")).intValue();

            // Get the interval for each heartbeat
            int hbInterval = Config.PRICEGEN_HEARTBEAT_INTERVAL;

            // For each record in the database, create an alarm to publish it when the time comes.
            Collection list = rtcHome.findAllTickers();
            logger.logp(Level.FINE, className, methodName, "Found " + list.size() + " tickers in the database.");
            Iterator iter = list.iterator();

            while(iter.hasNext())
            {
                RawTickerConfigLocal data = (RawTickerConfigLocal)iter.next();
                logger.logp(Level.FINE, className, methodName, "Creating alarm for " + data.getTicker(), new Object[]{data.getTickPeriod(),data.getStartPrice(),data.getPriceDeltaByTick(),data.getTickerDecisionRange(),data.getTickerUpTickRange(),data.getTickerDownTickRange()});

                PricePublisherAlarm ppa = new PricePublisherAlarm(data, topic);
                am.create(ppa, this, interval);
            }

            logger.logp(Level.FINE, className, methodName, "Starting heart beat alarm");
            Topic hbTopic = (Topic)ic.lookup("java:comp/env/jms/PriceHBTopic");

            // Start the HeartBeat alarm
            PriceHeartBeatAlarm hba = new PriceHeartBeatAlarm(hbInterval, hbTopic);
            am.create(hba, this, hbInterval);
        }
        catch(Exception e)
        {
            logger.logp(Level.SEVERE, className, methodName, "Exception Thrown", e);
        }
        finally
        {
            logger.exiting(className, methodName);
        }

        return true;
    }

    /**
     * Destroy the price publishers and all associated resources.
     * @return boolean True if the publishers were stopped.  False if
     * they were already stopped.
     */
    public boolean stopPublishers() throws Exception
    {
        String methodName = "stopPublishers";
        logger.entering(className, methodName);

        try
        {
            InitialContext ic = new InitialContext();

	        WorkManager wm = (WorkManager)ic.lookup("java:comp/env/wm/WorkManager");
        	AsynchScope as = wm.findAsynchScope(PriceGenConfig.AS_NAME);

            if (as == null)
            {
                // Already stopped.
                return false;
            }

            as.destroy();
        }
        catch(Exception e)
        {
            logger.logp(Level.SEVERE, className, methodName, "Exception Thrown", e);
            throw e;
        }
        finally
        {
            logger.exiting(className, methodName);
        }
        return true;
    }
	/**
	 * @see javax.ejb.SessionBean#ejbActivate()
	 */
	public void ejbActivate() throws EJBException, RemoteException
	{
	}

	/**
	 * @see javax.ejb.SessionBean#ejbPassivate()
	 */
	public void ejbPassivate() throws EJBException, RemoteException
	{
	}

	/**
	 * @see javax.ejb.SessionBean#ejbRemove()
	 */
	public void ejbRemove() throws EJBException, RemoteException
	{
	}

    /**
     * ejbCreate
     */
    public void ejbCreate() throws javax.ejb.CreateException {
    }

	/**
	 * @see javax.ejb.SessionBean#setSessionContext(SessionContext)
	 */
	public void setSessionContext(SessionContext arg0)
		throws EJBException, RemoteException
	{
        mySessionCtx = arg0;
	}

    /**
     * getSessionContext
     */
    public javax.ejb.SessionContext getSessionContext() {
        return mySessionCtx;
    }
}
