//
// "This program may be used, executed, copied, modified and distributed without royalty for the
// purpose of developing, using, marketing, or distributing."
//
// Product 5724-I63, 5724-H88, 5655-N02, 5733-W70,  (C) COPYRIGHT International Business Machines Corp., 2001,2006
// All Rights Reserved * Licensed Materials - Property of IBM
// US Government Users Restricted Rights - Use, duplication or disclosure
// restricted by GSA ADP Schedule Contract with IBM Corp.
//
package com.ibm.websphere.samples.asynchbeans.wstrader.tickstream;

import javax.ejb.DuplicateKeyException;
import javax.jms.*;
import javax.naming.*;

import java.util.Hashtable;
import java.util.Map;
import java.util.logging.*;

import com.ibm.websphere.asynchbeans.Alarm;
import com.ibm.websphere.asynchbeans.AsynchScope;
import com.ibm.websphere.asynchbeans.AsynchScopeEvents;
import com.ibm.websphere.asynchbeans.EventSource;
import com.ibm.websphere.asynchbeans.SubsystemMonitor;
import com.ibm.websphere.asynchbeans.SubsystemMonitorManager;
import com.ibm.websphere.asynchbeans.WorkException;
import com.ibm.websphere.asynchbeans.WorkManager;
import com.ibm.websphere.samples.asynchbeans.wstrader.*;

/**
 * This class holds the application state. This state tracks all application
 * wide state and resources. This state has an associated 'root' asynch bean
 * on a WorkManager. This object is stored on that AsynchBean using a property
 * allowing it to be retrieved easily at runtime.
 * We currently use a single TopicConnection for the applications. Any Thread/Works
 * make their own JMSSessions using this TopicConnection.
 */

public class StreamApplicationState implements AsynchScopeEvents
{
    static final Logger logger = Logger.getLogger(StreamApplicationState.class.getName());
    static final String className = StreamApplicationState.class.getName();
    static private final String PROP_APP_STATE = "appState";

    /**
     * The name of the Price Generator Subsystem Monitor
     */
    static public final String AS_SSM_PRICEGEN = "SSM_PriceGen";

	/**
	 * The WorkManager associate with the application.
	 */
	WorkManager workManager;
	/**
	 * This is the TopicConnectionFactory used by this application.
	 */
	TopicConnectionFactory tcf;

	/**
	 * This is the topic that prices are being published on. Individual tickers
	 * are identified by a particular property on the message. Selectors are
	 * used to receive prices for a particular ticker using this property.
	 */
	Topic priceTopic;
	/**
	 * This is the associated AsynchBean for this application. It is used
	 * for application wide SubsystemMonitors and Works etc.
	 */
	AsynchScope appAsynchScope;

    /**
     * This is the subsystem monitor for our Price Generator feed.
     */
    SubsystemMonitor priceGenSSM;

    /**
     * This map tracks all the JMS listeners in use by the application.
     */
    Map JMSListeners;

    /**
     * An alarm that send ping events to anyone who's registered as a listener to the
     * StreamHeartBeatEvents in it's
     */
    Alarm streamHBAlarm;

	/**
	 * Private constructor. Use the static getApplicationState method to
	 * create this.
	 */
	private StreamApplicationState(WorkManager wm)
		throws JMSException, NamingException
	{
        String methodName = "StreamApplicationState";
        logger.entering(className, methodName);

		try
		{
			workManager = wm;

			// the convention is to synch on the workManager when looking
			// up or creating the app asynch bean to avoid making two.
			synchronized(wm)
			{
				// we don't want to make two of these.
				if(wm.findAsynchScope(TickStreamConfig.AS_NAME) != null)
				{
					throw new java.lang.IllegalStateException("Stream App AsynchBean already exists");
				}
				try
				{
					appAsynchScope = wm.createAsynchScope(TickStreamConfig.AS_NAME);
				}
				catch(DuplicateKeyException e)
				{
					appAsynchScope = wm.findAsynchScope(TickStreamConfig.AS_NAME);
				}


				// we want all notifications from the app bean.
				// such as destroyed etc.
				//appBean.addNotificationListener(this, null, null);
				// The listener sequence is 9 so that we are called last. The others all use priority 5 (the default).
				// our destroy method cleans up the state and we only want this to happen
				// once the other listeners have reacted to us being destroyed.
                appAsynchScope.registerListener(appAsynchScope, this, EventSource.MAX_LISTENER_SEQUENCE_VALUE);

				// initialize the application state
				initialize();

				// store a reference to the app state in the app bean.
				// the static getApplicationState can then be used to retrieve
				// it later by getting the app asynch bean and then using this
				// property to get this state object.
				appAsynchScope.getPropertyMap().put(PROP_APP_STATE, this);

                SubsystemMonitorManager smm = appAsynchScope.getSubsystemMonitorManager();
                try
                {
                    priceGenSSM = smm.create(AS_SSM_PRICEGEN, Config.PRICEGEN_HEARTBEAT_INTERVAL, Config.PRICEGEN_HEARTBEAT_STALE, Config.PRICEGEN_HEARTBEAT_DEAD);
                    priceGenSSM.start();
                }
                catch(DuplicateKeyException e)
                {
                    logger.logp(Level.SEVERE, className, methodName, "constructor", e);
                    // not possible, it's a new asynch bean
                }

                // Create an alarm that will act as a heartbeat for the server.
                // Note that we are not preserving the context of our alarm since it
                // will not be using any J2EE context information.
                StreamHeartBeatAlarm streamHBAlarmListener = new StreamHeartBeatAlarm(appAsynchScope, Config.STREAMSERVER_HEARTBEAT_INTERVAL);
                streamHBAlarm = appAsynchScope.getAlarmManager().create(streamHBAlarmListener, null, Config.STREAMSERVER_HEARTBEAT_INTERVAL);
			}

			// This tracks the list of JMS Message listeners currently active in
			// the application. We basically have a single listener for each ticker.
			// Multiple subscribers to a ticker are supported using the EventSource
			// notification mechanism when messages arrive.
			JMSListeners = new Hashtable();
		}
		catch(JMSException e)
		{
			logger.logp(Level.SEVERE, className, methodName, "JMSException initializing application state", e);
			throw e;
		}
		catch(NamingException e)
		{
			logger.logp(Level.SEVERE, className, methodName, "NamingException initializing application state", e);
			throw e;
		}
		finally
		{
			logger.exiting(className, methodName);
		}
	}

	/**
	 * This is exposed to clients can register listeners on it to be notified when the price generator
	 * goes down or recovers.
	 */
	public SubsystemMonitor getPriceGeneratorSSM()
	{
		return priceGenSSM;
	}

    /**
     * Initialize this object
     */
	public void initialize()
		throws NamingException, JMSException
	{
        String methodName = "initialize";
        logger.entering(className, methodName);

		try
		{
			InitialContext ic = new InitialContext();

			// get a TopicConnection
	        tcf = (TopicConnectionFactory)ic.lookup("java:comp/env/jms/TCF");

            // get the Price Topic
	        priceTopic = (Topic) ic.lookup("java:comp/env/jms/PriceTopic");

		}
		catch(NamingException e)
		{
            logger.logp(Level.SEVERE, className, methodName, "NamingException during initialize", e);
			throw e;
		}
		finally
		{
			logger.exiting(className, methodName);
		}
	}

	/**
	 * This is the method used to get the ApplicationState object. It will
	 * create the state and initialize it if necessary.
	 */
	public static StreamApplicationState getApplicationState(WorkManager wm, boolean create)
		throws NamingException, JMSException
	{
		StreamApplicationState appState = null;
        String methodName = "getApplicationState";
        logger.entering(className, methodName);

		try
		{
			synchronized(wm)
			{
				AsynchScope appAS = wm.findAsynchScope(TickStreamConfig.AS_NAME);
				if(appAS == null)
				{
                    if(create)
                    {
                        logger.logp(Level.FINE, className, methodName, "Creating new Asynch Scope.");
    					appState = new StreamApplicationState(wm);
                    }
                    else
                    {
                        logger.logp(Level.FINE, className, methodName, "Asynch Scope not found.  Returning null.");
                        return null;
                    }
				}
				else
				{
					appState = (StreamApplicationState)appAS.getPropertyMap().get(PROP_APP_STATE);
				}
			}
		}
		finally
		{
			logger.exiting(className, methodName);
		}
		return appState;
	}

    /**
     * Destroy the AsynchScope where our StreamAppliationState is stored.  This
     * will in turn call the destroyed() event which will call _destroy()
     */
    public synchronized void destroy()
    {
        appAsynchScope.destroy();
    }

	/**
	 * This cleans up the application state and releases resources etc.
	 * This should only be called when the associated AsynchScope is destroyed.
	 */
	private void _destroy()
	{
        String methodName = "_destroy";
        logger.entering(className, methodName);

		try
		{
			synchronized(workManager)
			{
				appAsynchScope.getPropertyMap().clear();
                streamHBAlarm.cancel();
				appAsynchScope = null;
				tcf = null;
			}
		}
		finally
		{
			logger.exiting(className, methodName);
		}
	}

    /**
     * @see com.ibm.websphere.asynchbeans.AsynchScopeEvents#destroyed(AsynchScope)
     */
    public void destroyed(AsynchScope scope)
    {
        String methodName = "destroyed";
        logger.entering(className, methodName);

        _destroy();

        logger.exiting(className, methodName);
    }

    /**
     * @see com.ibm.websphere.asynchbeans.EventSourceEvents#listenerCountChanged(EventSource, int, int)
     */
    public void listenerCountChanged(
        EventSource es,
        int oldCount,
        int newCount)
    {
    }

    /**
     * @see com.ibm.websphere.asynchbeans.EventSourceEvents#listenerExceptionThrown(EventSource, Object, String, Throwable)
     */
    public void listenerExceptionThrown(
        EventSource es,
        Object listener,
        String methodName,
        Throwable exception)
    {
        logger.logp(Level.SEVERE, className, "listenerExceptionThrown", "Unexpected Exception ", exception);
    }

    /**
     * @see com.ibm.websphere.asynchbeans.EventSourceEvents#unexpectedException(EventSource, Object, Throwable)
     */
    public void unexpectedException(
        EventSource es,
        Object runnable,
        Throwable exception)
    {
        logger.logp(Level.SEVERE, className, "unexpectedException", "Unexpected Exception ", exception);
    }

    /**
     * Retrieve a new or existing client state.
     * @param clientID a String representation of a client.
     * @param createFlag If true and the client state doesn't exist, don't create one.  Return null.
     *                    If false, create and return a new StreamClientState object.
     * @return StreamClientState
     */
	public StreamClientState getClientState(String clientID, boolean createFlag)
	{
		StreamClientState clientState = null;
        String methodName = "getClientState";
        logger.entering(className, methodName);

		try
		{
			synchronized(appAsynchScope)
			{
				AsynchScope clientAS = appAsynchScope.findAsynchScope(clientID);
				if(clientAS == null && !createFlag)
				{
                    logger.logp(Level.FINE, className, methodName, "no client ab and didn't create one");
					return null;
				}
				if(clientAS == null)
				{
                    logger.logp(Level.FINE, className, methodName, "no client ab and creating one");
					clientState = StreamClientState.getClientState(this, clientID, createFlag);
				}
			}
		}
		finally
		{
			logger.exiting(className, methodName);
		}
		return clientState;
	}

	/**
	 * We could have given each attach client its own set of dynamically started
	 * JMS listeners. This won't scale at all however. Instead, clients share jms listeners
	 * for the same tickers. This method is used by clients to find a listener
	 * for their ticker. We return an existing ticker if one was already in use
	 * by other clients. We create a new one if this client is the first client
	 * to ask for messages for that ticker. A Map keyed on ticker is used to
	 * track the listeners so that they can be shared in this manner.<p>
	 * Clients then add a listener to the EventSource associated with the listener
	 * and will then get an event when a message is received.
	 * @param ticker We need a listener for this ticker.
	 * @return The listener for the ticker.
	 */
    public synchronized TickerSubscriber getTickerSubscriber(String ticker)
        throws WorkException, JMSException, NamingException
    {
        String methodName = "getTickerSubscriber";
        logger.entering(className, methodName);
        TickerSubscriber tickSub = null;
    	try
    	{
    		// if we already had one then return it from the Map.
	        if (JMSListeners.containsKey(ticker))
	        {
                logger.logp(Level.FINE, className, methodName, "found existing ticker", ticker);
	            tickSub = (TickerSubscriber) JMSListeners.get(ticker);
	        }
	        // otherwise, create a new one.
	        else
	        {
                logger.logp(Level.FINE, className, methodName, "create new ticker", ticker);
	            tickSub = new TickerSubscriber(this, workManager, tcf, priceTopic, ticker);
	            JMSListeners.put(ticker, tickSub);
	        }
    	}
    	finally
    	{
			logger.exiting(className, methodName);
    	}
        return tickSub;

    }

	/**
	 * This is called by TickerSubscriber when it notices that there are no
	 * more listeners attached to it. For example, there are no clients with an interest
	 * in that ticker anymore.
	 */
	public synchronized void removeTickerSubscriber(String ticker)
	{
        String methodName = "removeTickerSubscriber";
        logger.entering(className, methodName);
		try
		{
			if(JMSListeners.containsKey(ticker))
			{
                logger.logp(Level.FINE, className, methodName, "removeTickerSubscriber", ticker);
				JMSListeners.remove(ticker);
			}
			else
			{
                logger.logp(Level.FINE, className, methodName, "unknown ticker", ticker);
			}
		}
		finally
		{
			logger.exiting(className, methodName);
		}
	}

    /**
     * This is called whenever a heart beat message is received from the price generator.
     * We simply delegate to the ping on the SSM tracking the health of the price generator.
     */
    public void handlePriceGenPing()
    {
        logger.logp(Level.FINEST, className, "handlePriceGenPing", "ping detected");
        priceGenSSM.ping();
    }

    /**
     * Retrieve the Asynchronous Scope associated with this application state.
     */
    public AsynchScope getAsynchScope()
    {
        return appAsynchScope;
    }
}
