//
// "This program may be used, executed, copied, modified and distributed without royalty for the
// purpose of developing, using, marketing, or distributing."
//
// Product 5724i63,  (C) COPYRIGHT International Business Machines Corp., 2001,2004
// All Rights Reserved * Licensed Materials - Property of IBM
// US Government Users Restricted Rights - Use, duplication or disclosure
// restricted by GSA ADP Schedule Contract with IBM Corp.
//
package com.ibm.websphere.samples.asynchbeans.wstrader.tickstream;

import java.util.Hashtable;
import java.util.Map;

import javax.ejb.DuplicateKeyException;
import javax.management.Notification;
import javax.management.NotificationListener;

import java.util.logging.*;
import com.ibm.websphere.asynchbeans.AsynchScope;
import com.ibm.websphere.asynchbeans.AsynchScopeEvents;
import com.ibm.websphere.asynchbeans.EventSource;
import com.ibm.websphere.asynchbeans.SubsystemMonitor;
import com.ibm.websphere.asynchbeans.SubsystemMonitorEvents;
import com.ibm.websphere.asynchbeans.SubsystemMonitorManager;
import com.ibm.websphere.asynchbeans.WorkManager;
import com.ibm.websphere.samples.asynchbeans.wstrader.Config;

/**
 * Contains all logic for maintaining the state of a client (login id).
 * Each ClientState has an associated Asynch Scope which is a child
 * of the main Application Asynch Scope or StreamApplicationState.
 */
public class StreamClientState implements AsynchScopeEvents, SubsystemMonitorEvents
{
    static final Logger logger = Logger.getLogger(StreamClientState.class.getName());
    static final String className = StreamClientState.class.getName();

	final private static String PROP_CLIENT_STATE = "STATE_CLIENT";
    final public static String SSM_CLIENT = "SSM_CLIENT";

	AsynchScope clientAsynchScope;
	WorkManager workManager;
	StreamApplicationState appState;
	String clientID;
	SubsystemMonitor clientSSM;
	boolean clientIsPinging;


	/**
     * This is a map of all active streaming connections for this client.
     */
	Map activeStreamers;

    /**
     * StreamClientState
     * @param appState the StreamApplicationState associated with this app
     * @param clientID the id of the client that this object is managing the state of.
     */
	private StreamClientState(StreamApplicationState appState, String clientID)
	{
        String methodName = "StreamClientState";
        logger.entering(className, methodName);

		try
		{
			this.clientID = clientID;
			this.appState = appState;
			clientIsPinging = false;

            // Find the singleton Application State Asynch Bean and
            // store a reference to this Client State object inside the bean in its property map.
			AsynchScope appAS = appState.getAsynchScope();

			synchronized(appAS)
			{
				clientAsynchScope = appAS.findAsynchScope(clientID);
				if(clientAsynchScope != null)
				{
                    logger.logp(Level.FINE, className, methodName, "client ab already exists.");
					throw new IllegalStateException("Client Asynch Bean already exists");
				}
				try
				{
					clientAsynchScope = appAS.createAsynchScope(clientID);
				}
				catch(DuplicateKeyException e)
				{
					clientAsynchScope = appAS.findAsynchScope(clientID);
				}
				clientAsynchScope.getPropertyMap().put(PROP_CLIENT_STATE, this);

                // Register ourself as a listener for any Asynch Bean notifications
                // we use listener order sequence 9 to make sure that we are called last. This is important
                // because the client state destroy implementation clears a lot of state that may be
                // needed by other listeners. So, we want it to run last to prevent the others listeners
                // trying to access destroyed state.
                clientAsynchScope.registerListener(clientAsynchScope, this, EventSource.MAX_LISTENER_SEQUENCE_VALUE);

                // Create a Subsystem Monitor to receive all Client Heartbeats
                // Register ourself as a listener for all Subystem Monitor notifications (dead/stale etc)
				SubsystemMonitorManager smm = clientAsynchScope.getSubsystemMonitorManager();
				try
				{
					clientSSM = smm.create(SSM_CLIENT, Config.CLIENT_HEARTBEAT_INTERVAL, Config.CLIENT_HEARTBEAT_STALE, Config.CLIENT_HEARTBEAT_DEAD);
                    clientAsynchScope.registerListener(clientSSM, this);
					clientSSM.start();
				}
				catch(DuplicateKeyException e)
				{
                    logger.logp(Level.SEVERE, className, methodName, "DuplicateKeyException " + e);
					// not possible, it's a new asynch bean
				}
			}

            // Allocate our list of active streamers (e.g. "DEFAULT")
			activeStreamers = new Hashtable();
		}
		finally
		{
			logger.exiting(className, methodName);
		}
	}

    /**
     * Create a StreamClientState instance for a specific client id and application state.
     * @param appState the associated StreamApplicationState
     * @param clientID the client id to associate with the StreamClientState.
     * @param createFlag when true, create a new StreamClientState when one doesn't already exist.
     */
	public static StreamClientState getClientState(StreamApplicationState appState, String clientID, boolean createFlag)
	{
        String methodName = "getClientState";
        logger.entering(className, methodName);
		StreamClientState clientState = null;
		try
		{
			AsynchScope appAS = appState.getAsynchScope();

			if(appAS != null)
			{
				// it can be null after we are destroyed.
				synchronized(appAS)
				{
					AsynchScope clientAsynchScope = appAS.findAsynchScope(clientID);
					if(clientAsynchScope == null && !createFlag)
					{
                        logger.logp(Level.FINE, className, methodName, "not found and cannot create", clientID);
						return null;
					}
					if(clientAsynchScope == null)
					{
                        logger.logp(Level.FINE, className, methodName, "not found and create new one", clientID);
						clientState = new StreamClientState(appState, clientID);
					}
					else
					{
                        logger.logp(Level.FINE, className, methodName, "found existing one", clientID);
						clientState = (StreamClientState)clientAsynchScope.getPropertyMap().get(PROP_CLIENT_STATE);
					}
				}
			}
		}
		finally
		{
			logger.exiting(className, methodName);
		}
		return clientState;
	}

    /**
     * Destroy this instance.
     */
    public void destroy()
    {
        logger.logp(Level.FINE, className, "destroy", "Destroying Client State: " + clientID);
        clientAsynchScope.destroy();
    }

    /**
     * Internal destroy of this instance.
     */
	private void _destroy()
	{
        String methodName = "_destroy";
        logger.entering(className, methodName);
        logger.logp(Level.FINE, className, methodName, "Destroying Client State: ", clientID);
		try
		{
			workManager = null;
			clientAsynchScope = null;
			appState = null;
		}
		finally
		{
			logger.exiting(className, methodName);
		}
	}

    /**
     * @see com.ibm.websphere.asynchbeans.AsynchScopeEvents#destroyed(AsynchScope)
     */
    public void destroyed(AsynchScope scope)
    {
        // Destroy this
        if(scope == clientAsynchScope)
        {
            _destroy();
        }
    }

    /**
     * @see com.ibm.websphere.asynchbeans.SubsystemMonitorEvents#destroyed(SubsystemMonitor)
     */
    public void destroyed(SubsystemMonitor ssm)
    {
    }

    /**
     * @see com.ibm.websphere.asynchbeans.SubsystemMonitorEvents#subsystemIsDead(SubsystemMonitor)
     */
    public void subsystemIsDead(SubsystemMonitor ssm)
    {
        // Our is no longer responding, so destroy it's scope
        if (ssm == clientSSM)
        {
            clientAsynchScope.destroy();
        }
    }

    /**
     * @see com.ibm.websphere.asynchbeans.SubsystemMonitorEvents#subsystemIsFresh(SubsystemMonitor)
     */
    public void subsystemIsFresh(SubsystemMonitor ssm)
    {
    }

    /**
     * @see com.ibm.websphere.asynchbeans.SubsystemMonitorEvents#subsystemIsStale(SubsystemMonitor)
     */
    public void subsystemIsStale(SubsystemMonitor ssm)
    {
    }

    /**
     * @see com.ibm.websphere.asynchbeans.EventSourceEvents#listenerCountChanged(EventSource, int, int)
     */
    public void listenerCountChanged(
        EventSource es,
        int oldCount,
        int newCount)
    {
    }

    /**
     * @see com.ibm.websphere.asynchbeans.EventSourceEvents#listenerExceptionThrown(EventSource, Object, String, Throwable)
     */
    public void listenerExceptionThrown(
        EventSource es,
        Object listener,
        String methodName,
        Throwable exception)
    {
        logger.logp(Level.SEVERE, className, "listenerExceptionThrown", "Exception: " + exception);
    }

    /**
     * @see com.ibm.websphere.asynchbeans.EventSourceEvents#unexpectedException(EventSource, Object, Throwable)
     */
    public void unexpectedException(
        EventSource es,
        Object runnable,
        Throwable exception)
    {
        logger.logp(Level.SEVERE, className, "unexpectedException", "Exception: " + exception);
    }



    /**
     * Handles pings from the Ping servlet for a given client
     */
	public void handleClientPing()
	{
        logger.logp(Level.FINEST, className, "handleClientPing", "ping detected");
		clientSSM.ping();
	}

	/**
     * Get the ID of the client that is associated with this Client State
     */
    public String getClientID()
	{
		return clientID;
	}

    /**
     * Get the client Asynch Scope that is associated with this Client State
     */
	public AsynchScope getClientAsynchScope()
	{
		return clientAsynchScope;
	}
	/**
	 * Gets the appState.
	 * @return Returns a StreamApplicationState
	 */
	public StreamApplicationState getApplicationState() {
		return appState;
	}

	/**
	 * Sets the appState.
	 * @param appState The appState to set
	 */
	public void setApplicationState(StreamApplicationState appState) {
		this.appState = appState;
	}

	/**
	 * A client can have multiple streaming connections concurrently. Each
	 * one is named uniquely and registered here so that it can be found
	 * easily using it's name.
     * @param streamer The streamer to register.
	 */
	public synchronized void registerStreamer(StreamListener streamer)
	{
        logger.logp(Level.FINE, className, "registerStreamer", "registering " + streamer.getName());
		activeStreamers.put(streamer.getName(), streamer);
	}

    /**
     * Unregister a streamer from the client state.
     * @param streamer The streamer to unregister.
     */
	public synchronized void unregisterStreamer(StreamListener streamer)
	{
		if(activeStreamers.containsKey(streamer.getName()))
		{
            logger.logp(Level.FINE, className, "unregisterStreamer", "unregistering", streamer.getName());
			activeStreamers.remove(streamer.getName());
		}
		else
		{
            logger.logp(Level.FINE, className, "unregisterStreamer", "no such streamer", streamer.getName());
		}
	}

    /**
     * Find a streamer by it's name
     * @param name The streamer name.
     */
	public synchronized StreamListener findStreamer(String name)
	{
        String methodName = "findStreamer";
        logger.entering(className, methodName);
		try
		{
			StreamListener rc = (StreamListener)activeStreamers.get(name);
			if(rc == null)
                logger.logp(Level.FINE, className, methodName, "stream not found:", name);
			return rc;
		}
		finally
		{
			logger.exiting(className, methodName);
		}
	}

    public String toString()
    {
        return "<ClientState for " + clientID + ">";
    }
}
