//
// "This program may be used, executed, copied, modified and distributed without royalty for the
// purpose of developing, using, marketing, or distributing."
//
// Product 5724-I63, 5724-H88, 5655-N02, 5733-W70,  (C) COPYRIGHT International Business Machines Corp., 2001,2006
// All Rights Reserved * Licensed Materials - Property of IBM
// US Government Users Restricted Rights - Use, duplication or disclosure
// restricted by GSA ADP Schedule Contract with IBM Corp.
//
package com.ibm.websphere.samples.asynchbeans.wstrader.tickstream;

import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.StringTokenizer;

import javax.jms.JMSException;
import javax.jms.Message;
import javax.naming.NamingException;

import java.util.logging.*;
import com.ibm.websphere.asynchbeans.AsynchScope;
import com.ibm.websphere.asynchbeans.AsynchScopeEvents;
import com.ibm.websphere.asynchbeans.EventSource;
import com.ibm.websphere.asynchbeans.SubsystemMonitor;
import com.ibm.websphere.asynchbeans.SubsystemMonitorEvents;
import com.ibm.websphere.asynchbeans.WorkException;
import com.ibm.websphere.samples.asynchbeans.wstrader.ClientInstruction;

/**
 * This is created the by stream servlet. It's job is to listen for events that
 * need to be pushed over HTTP to the applet and push them.  Each applet client
 * will have a StreamListener associated with it.  When the applet disconnects, the 
 * listener will remain, but the PrintWriter (associated with the response object of the
 * Servlet) will become invalid.  The client can then reconnect, look-up this listener instance
 * and retrieve various settings, such at the Ticker symbols that were previously configured.
 * 
 * If the client manually logs off, then this listener should be explicitly destroyed.
 */
public class StreamListener implements AsynchScopeEvents, SubsystemMonitorEvents, TickerUpdateEvents, StreamHeartBeatEvents
{
    static final Logger logger = Logger.getLogger(StreamListener.class.getName());
    static final String className = StreamListener.class.getName();
	/**
	 * The name of the streamer.
	 */
    String name;

    /**
     * The buffer used to stream the output to.  Someone else will
     * pull the data out of the other end of the buffer.
     */
    CurrentQueue buffer;
    
    /**
     * The client state (login) associated with this stream.
     */
    StreamClientState clientState;

    /**
     * A reference to the application state.
     */
    StreamApplicationState appState;
    
    /**
     * The the servlet's PrintWriter.  If this becomes invalid, it will
     * automatically be set to null to indicate that the servlet is no longer valid.
     */
    PrintWriter pw;

    /**
     * Every JMS source that we're registered with goes here. Each ticker
     * we're interested in has a JMS source and it's stored in this
     * map using the ticker name as a key.
     */
    Map registeredSources;

    /**
     * The ticker symbols that the client has subscribed to.
     */
	String subscribedTickers;

	/**
	 * This is false when we don't want to send quotes anymore.
	 */
	boolean sendingEvents;
    
    /**
     * The AsynchScope that is associated with the client.
     */
    AsynchScope clientAsynchScope;
    
    /**
     * Create a new Servlet Listener
     * @param cstate the client state that this stream should be registered with.  A client may have more than one
     * stream instance.
     * @param pw the servlet's response object's PrintWriter.
     * @param name the identity of this stream.  The client can lookup this stream by name any time.
     */
    public StreamListener(StreamClientState cstate, PrintWriter pw, String name)
    {
        String methodName = "StreamListener";
        logger.entering(className, methodName);
    	try
    	{
	    	this.name = name;
	    	this.pw = pw;
	    	sendingEvents = true;
	    	clientState = cstate;
	    	clientAsynchScope = clientState.getClientAsynchScope();

            appState = clientState.getApplicationState();
            registeredSources = new Hashtable();

            // Create a special buffer that will prevent slow clients from blocking
            // the JMS Topic Subscriber.
	        buffer = new CurrentQueue(15);

            // Notify this object instance of any Client AsynchBean events (such as destroy)
            clientAsynchScope.registerListener(clientAsynchScope,this);

            // Add this Streamer/Listener to this Client's list of active streamers.
	        clientState.registerStreamer(this);

            // Register ourself as a listener on the price generator SSM
            clientAsynchScope.registerListener(appState.getPriceGeneratorSSM(), this);
            
            // this is to get the stream heart beat events. This is basically an alarm that causes this
            // listener to send a heart beat to the applet to let it know the server is alive.
            clientAsynchScope.registerListener(appState.getAsynchScope(), this);

    	}
    	finally
    	{
    		logger.exiting(className, methodName);
    	}
    }

	/**
	 * This returns the name of this streamer.
	 */
	public String getName()
	{
		return name;
	}

	/**
	 * Wakes all all routines waking on the PrintWriter. This should
	 * only be the servlet on which the streaming request was received.
	 */
	public void wakeServlet()
	{
        logger.logp(Level.FINE, className, "wakeServlet", "");
        if(pw!=null)
        {
            // If the PrintWriter is null, then the servlet response object is gone.
    		synchronized(pw)
    		{
    			pw.notifyAll();
    		}
        }
	}

    /**
     * Stop this streamer from listening to all ticker subscribers.
     */
	private void listenerFinished(boolean duringDestroy)
	{
        String methodName = "listenerFinished";
        logger.entering(className, methodName);

        // This listener is also handling the ticker messages, so unregister it from
        // all of the ticker subscribers.
        // Remove all ticker listeners

        try {
            logger.logp(Level.FINE, className, methodName, "Cleaning up for Client,streamer:", new Object[]{clientState.getClientID(), name});

        	// Stop sending messages.
        	setSendingEvents(false);

            // Close the queue
            buffer.close();

        	// Unsubscribe from all tickers.
        	updateTickers("", duringDestroy);

        	clientState.unregisterStreamer(this);
        	if(!duringDestroy)
        	{
        		// no need to do this during a destroy and it's done
        		// automatically. If this method is called during the
        		// AsynchScopeEvents.destroy method then we cannot call this
        		// as the AsynchScope is marked as destroyed. If we did then
        		// we'd get an IllegalStateException
            	clientAsynchScope.unregisterListener(clientAsynchScope,this);

                // Unregister our listener from PriceGenerator subsystem monitor. 
                clientAsynchScope.unregisterListener(appState.getPriceGeneratorSSM(), this);
                
                // Unregister our listener from the heart beat generator. 
                clientAsynchScope.unregisterListener(appState.getAsynchScope(), this);
            }

        	// If a servlet thread is blocked waiting for the stream to be
        	// finished then notify it so that it can wake up and return
        	// to the servlet engine thread pool.
        	wakeServlet();
        }
        catch(IllegalStateException e)
        {
        	// possible if we're called as AsynchScope is being destroyed.
			logger.logp(Level.SEVERE, className, methodName, "Exception: " + e);
        }
        catch (Exception e)
        {
			logger.logp(Level.SEVERE, className, methodName, "Exception: " + e);
        }
        finally
        {
        	logger.exiting(className, methodName);
        }
	}

	/**
    /**
     * This is only called when either the client logs out or the SSM associated with the clienyt
     * detected that the client is not responsive. When these events happen then this gets called
     * and we clean up the state for the client and wake up the stream servlet to allow it to be reused.
     * @see com.ibm.websphere.asynchbeans.AsynchScopeEvents#destroyed(AsynchScope)
     */
    public void destroyed(AsynchScope scope)
    {
        if (scope == clientState.getClientAsynchScope())
        {
            logger.logp(Level.FINE, className, "destroyed", "Destroying Listener for client and stream:", new Object[]{name, clientState.getClass()});
            listenerFinished(true);
        }
    }

    /**
     * @see com.ibm.websphere.asynchbeans.SubsystemMonitorEvents#destroyed(SubsystemMonitor)
     */
    public void destroyed(SubsystemMonitor ssm)
    {
    }

    /**
     * This gets called when the SSM associated with the price generator detects the price generator
     * has failed or been stopped.
     * @see com.ibm.websphere.asynchbeans.SubsystemMonitorEvents#subsystemIsDead(SubsystemMonitor)
     */
    public void subsystemIsDead(SubsystemMonitor ssm)
    {
        // Price Generator Notification
        if(ssm == appState.getPriceGeneratorSSM())
        {
            sendMessage(ClientInstruction.createStatusResponse(
                ClientInstruction.RESPONSE_STATUS_SOURCE_PRICEGEN,
                ClientInstruction.RESPONSE_STATUS_VALUE_DEAD).getTextResponse());
        }
    }

    /**
     * This is called when the SSM associated with the price generator detects the price generator
     * has done a lazerus and recovered from a failure.
     * @see com.ibm.websphere.asynchbeans.SubsystemMonitorEvents#subsystemIsFresh(SubsystemMonitor)
     */
    public void subsystemIsFresh(SubsystemMonitor ssm)
    {
        // Price Generator Notification
        if(ssm == appState.getPriceGeneratorSSM())
        {
            sendMessage(ClientInstruction.createStatusResponse(
                ClientInstruction.RESPONSE_STATUS_SOURCE_PRICEGEN,
                ClientInstruction.RESPONSE_STATUS_VALUE_FRESH).getTextResponse());
        }
    }

    /**
     * This is called when the SSM associated with the price generator detects that the price generator
     * may fail. For example, it is late delivering heart beats to this server.
     * @see com.ibm.websphere.asynchbeans.SubsystemMonitorEvents#subsystemIsStale(SubsystemMonitor)
     */
    public void subsystemIsStale(SubsystemMonitor ssm)
    {
        // Price Generator Notification
        if(ssm == appState.getPriceGeneratorSSM())
        {
            sendMessage(ClientInstruction.createStatusResponse(
                ClientInstruction.RESPONSE_STATUS_SOURCE_PRICEGEN,
                ClientInstruction.RESPONSE_STATUS_VALUE_STALE).getTextResponse());
        }
    }

    /**
     * @see com.ibm.websphere.asynchbeans.EventSourceEvents#listenerCountChanged(EventSource, int, int)
     */
    public void listenerCountChanged(
        EventSource es,
        int oldCount,
        int newCount)
    {
    }

    /**
     * @see com.ibm.websphere.asynchbeans.EventSourceEvents#listenerExceptionThrown(EventSource, Object, String, Throwable)
     */
    public void listenerExceptionThrown(
        EventSource es,
        Object listener,
        String methodName,
        Throwable exception)
    {
        logger.logp(Level.SEVERE, className, "listenerExceptionThrown", "Unexpected Exception ", exception);
    }

    /**
     * @see com.ibm.websphere.asynchbeans.EventSourceEvents#unexpectedException(EventSource, Object, Throwable)
     */
    public void unexpectedException(
        EventSource es,
        Object runnable,
        Throwable exception)
    {
        logger.logp(Level.SEVERE, className, "unexpectedException", "Unexpected Exception ", exception);
    }

    /**
     * @see com.ibm.websphere.samples.asynchbeans.wstrader.tickstream.TickerUpdateEvents#tickerStopped(TickerSubscriber)
     */
	public void tickerStopped(TickerSubscriber subscriber)
	{
	}

    /**
     * This is called whenever a new message is received for a ticker that we're
     * interested in. We verify that we want the message and then queue it in the
     * FIFO associated with this client.<p>
     * The message is then sent by the sendMessages method on another thread. We return
     * quickly here to avoid holding up the message receiver.
     * @see com.ibm.websphere.samples.asynchbeans.wstrader.tickstream.TickerUpdateEvents#tickerUpdated(TickerSubscriber, Message)
     */
    public void tickerUpdated(TickerSubscriber subscriber, Message msg)
    {
    	try
    	{
    		String resp = null;
    		// need to synchronized because of registered sources.
    		synchronized(this)
    		{
	    		if(registeredSources.containsKey(subscriber.getTicker()))
	    		{
			        ClientInstruction clientInstruction = ClientInstruction.createTickerResponse(subscriber.getTicker(), msg.getStringProperty("type"), msg.getDoubleProperty("price"), msg.getDoubleProperty("bid"), msg.getDoubleProperty("ask"),msg.getStringProperty("time"));
			        resp = clientInstruction.getTextResponse();

	    		}
    		}
	        buffer.enQueue(resp);
    	}
    	catch(JMSException exception)
    	{
            logger.logp(Level.SEVERE, className, "tickerUpdated", "JMSException extracting price information from message", exception);
    	}
    }

    /**
     * Handle pings from the Stream HeartBeat Alarm
     * @see com.ibm.websphere.samples.asynchbeans.wstrader.tickstream.StreamHeartBeatEvents#handleHeartBeat(ClientInstruction)
     */
    public void handleHeartBeat(ClientInstruction clientInstruction)
    {
        sendMessage(clientInstruction.getTextResponse());
    }


    /**
     * Send all ticker update messages from our buffer to the associated
     * servlet's PrintWriter.  Continue to process messages as they become
     * available until getSendingEvents() becomes false or there is
     * an error writing to the PrintWriter.
     */
    public void sendMessages()
    {
        String methodName = "sendMessages";
        logger.entering(className, methodName);

        try
        {
            // Send the current state of our Price Generator engine
            // Lookup the PriceGen SSM and get it's current state and fire
            // the appropriate notification.
            SubsystemMonitor smPriceGen = appState.getAsynchScope().getSubsystemMonitorManager().find(StreamApplicationState.AS_SSM_PRICEGEN);
            int lastPriceGenState = SubsystemMonitor.SUBSYSTEM_DEAD;
            if (smPriceGen != null)
            {
                lastPriceGenState = smPriceGen.getCurrentState();
            }

            switch(lastPriceGenState)
            {
            	case SubsystemMonitor.SUBSYSTEM_DEAD:
                	subsystemIsDead(smPriceGen);
                	break;
                case SubsystemMonitor.SUBSYSTEM_STALE:
	                subsystemIsStale(smPriceGen);
	                break;
	            case SubsystemMonitor.SUBSYSTEM_FRESH:
                	subsystemIsFresh(smPriceGen);
                	break;
            }

            // Our message loop
	    	while(!buffer.isClosed && getSendingEvents())
	    	{
	    		// wait for 2 seconds maximum
                Object msg = buffer.deQueue(2000);
                // msg is null if deQueue times out or is closed.
                if(msg != null && sendMessage(msg.toString()))
                {
                    logger.logp(Level.FINE, className, methodName, "Ending Message Loop.  IO Error writing to client.");
                    break;
                }
	        }
        }
        finally
        {
            logger.exiting(className, methodName);
        }
    }

    /**
     * Send a message to our Servlet's PrintWriter.
     * @param msg the text to send
     * @return boolean true if there is an error
     */
    private boolean sendMessage(String msg)
    {
        String methodName = "sendMessage";
        logger.entering(className, methodName);
    	boolean hadProblem = true;
        
        // If the PrintWriter is null, then the servlet response object is gone.
        if(pw!=null)
        {
            // Synchronize all writes to the servlet stream.
            synchronized(pw)
            {
                logger.logp(Level.FINE, className, methodName, "sending Msg");
                pw.print(msg);
                logger.logp(Level.FINE, className, methodName, "sent Msg");
                hadProblem = pw.checkError();
            }
        }
        
        if(hadProblem)
        {
            pw=null;
        }
        
    	logger.exiting(className, methodName);
        return hadProblem;
    }

    /**
     * Convert a comma-delimited list of ticker symbols to a Set.
     * @param tickers comma-delimited list of ticker symbols
     * @return Set of Strings
     */
	Set parseTickerList(String tickers)
	{
    	Set newTickers = new HashSet();

        StringTokenizer stTickers = new StringTokenizer(tickers, ",");
        while(stTickers.hasMoreTokens())
        {
            String curTicker = stTickers.nextToken();
            newTickers.add(curTicker);
        }
        return newTickers;
	}

	/**
	 * This takes a new list of tickers to subscribe to.
	 * We remove any listeners from before which are not specified now
	 * and add new listeners for tickers we were not subscribed to before.
	 * @param tickers The new list of tickers for this client, can be empty meaning none.
	 * @param duringDestroy If true then the client has logged out or timed out and the AsynchScope for the client is being destroyed.
	 */
    public synchronized void updateTickers(String tickers, boolean duringDestroy)
    	throws NamingException, WorkException, JMSException
    {
        String methodName = "updateTickers";
        logger.entering(className, methodName);
    	try
    	{
	        if (tickers != null)
	        {
				subscribedTickers = tickers;

	        	// First, split up the comma delimited ticker list
	        	// and put entries in a Set.
	        	Set newTickers = parseTickerList(tickers);

	            // Now remove inactive subscriptions and remove
	            // entries for those tickers from the JMSSources
	            // Map.

	            Iterator iter = registeredSources.values().iterator();
	            ArrayList sourcesToRemove = new ArrayList();

	            while(iter.hasNext())
	            {
	            	TickerSubscriber sub = (TickerSubscriber)iter.next();
	            	if(!newTickers.contains(sub.getTicker()))
	            	{
                        logger.logp(Level.FINE, className, methodName, "Removing subscription", sub.getTicker());
            			if(!duringDestroy)
            			{
            				// only do this if this method was called as a result of a ticker update request. If
            				// this is called because the client has been destroyed then this unregister is
            				// done automatically when the ClientAsynchScope was destroyed.
                        	clientState.getClientAsynchScope().unregisterListener(sub.getEventSource(), this);
            			}
	            		sourcesToRemove.add(sub.getTicker());
	            	}
	            }

				// Do it after we're finished with iterator to avoid concurrent
				// modification exceptions.
				for(int i = 0; i < sourcesToRemove.size(); ++i)
				{
					registeredSources.remove(sourcesToRemove.get(i));
				}

	            // Now subscribe to any new tickers we were not subscribed
	            // to before.
	            iter = newTickers.iterator();
	            while(iter.hasNext())
	            {
	            	String ticker = (String)iter.next();
	            	if(registeredSources.containsKey(ticker))
	            	{
	            		// already subscribed, do nothing.
	            	}
	            	else
	            	{
		                TickerSubscriber curTickSub = clientState.getApplicationState().getTickerSubscriber(ticker);
                        logger.logp(Level.FINE, className, methodName, "Adding subscription", curTickSub.getTicker());

                        clientState.getClientAsynchScope().registerListener(curTickSub.getEventSource(), this);
		                registeredSources.put(ticker, curTickSub);
	            	}
	            }
	        }
    	}
    	finally
    	{
    		logger.exiting(className, methodName);
    	}
    }

	/**
	 * Gets the sendingEvents.
	 * @return Returns a boolean
	 */
	private synchronized boolean getSendingEvents() {
		return sendingEvents;
	}

	/**
	 * Sets the sendingEvents.
	 * @param sendingEvents The sendingEvents to set
	 */
	private synchronized void setSendingEvents(boolean sendingEvents) {
		this.sendingEvents = sendingEvents;
	}
	/**
	 * Returns the subscribedTickers.
	 * @return String
	 */
	public String getSubscribedTickers() {
		return subscribedTickers;
	}

}
