//
// "This program may be used, executed, copied, modified and distributed without royalty for the
// purpose of developing, using, marketing, or distributing."
//
// Product 5724i63,  (C) COPYRIGHT International Business Machines Corp., 2001,2004
// All Rights Reserved * Licensed Materials - Property of IBM
// US Government Users Restricted Rights - Use, duplication or disclosure
// restricted by GSA ADP Schedule Contract with IBM Corp.
//
package com.ibm.websphere.samples.asynchbeans.wstrader.tickstream;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Hashtable;

import javax.jms.*;
import javax.management.*;
import javax.naming.*;
import java.util.logging.*;

import com.ibm.websphere.asynchbeans.*;
import com.ibm.websphere.samples.asynchbeans.wstrader.*;

/**
 * This is a Work that is used to dynamically subscribe to a Topic and notify
 * listeners that are colocated in this JVM whenever a message is received
 * on that topic.<p>
 * It is created whenever a client that is attached needs to subscribe to a particular
 * stock ticker. A Map of these TickerSubscribers is kept in StreamApplicationState. If
 * a Tickersubscriber for the required ticker doesn't exist then we create a new
 * one for that ticker and add it to the Map. The client then adds a listener to
 * the EventSource associated with the TickerSubscriber object. The tickerUpdated
 * method on the client listener will be called whenever a message arrives on that
 * topic.<p>
 * The TickerSubscriber should only exist as long as their are interested clients
 * attached to its EventSource. Once the number of listeners falls to 1 then the
 * TickerSubscriber will terminate and release the JMS and thread resources.<p>
 */
public class TickerSubscriber implements Work, EventSourceEvents
{
    static final Logger logger = Logger.getLogger(TickerSubscriber.class.getName());
    static final String className = TickerSubscriber.class.getName();

	/**
	 * This is the application asynch state object.
	 */
	StreamApplicationState appState;

	/**
	 * This is an EventSource used by this class to fire events when messages arrive.
	 */
	EventSource eventSource;
	/**
	 * This proxy is used to fire events of this type to listeners of eventSource.
	 */
	TickerUpdateEvents tickerUpdateProxy;

	/**
	 * This is the topic connection factory that published prices
	 * are associated with.
	 */
	TopicConnectionFactory tcf;
	/**
	 * This is the topic that is receiving the messages.
	 */
	Topic topic;
	/**
	 * This is the stock ticker that is used as in our selector.
	 */
    String ticker;

    /**
     * This is the WorkItem for this object which is a Work.
     */
    WorkItem me;

    boolean released = false;

	public String getTicker()
	{
		return ticker;
	}

    /**
     * This creates a new TickerSubscriber. It receives the owner of this object,
     * appState and the ticker that needs to be subscribed to. It creates an
     * EventSource for this ticker and adds itself as a listener so that it
     * can detect when the listener count falls to 1, that it, itself. It then
     * starts a demon Work to subscribe to the topic and when a message is received
     * then it will fire an event to any listeners on EventSource that implement
     * TickerUpdateEvents.
     * @param wm the main WorkManager for the application.
     * @param selector the ticker that this TickerSubscriber is subscribed to.
     */
    public TickerSubscriber(StreamApplicationState appState, WorkManager wm, TopicConnectionFactory tcf, Topic topic, String selector)
        throws JMSException, WorkException, NamingException
    {
        String methodName = "TickerSubscriber";
        logger.entering(className, methodName);
    	try
    	{
	    	this.appState = appState;
	    	this.tcf = tcf;
	    	this.topic = topic;
	        this.ticker = selector;

	        // create the EventSource for the interested clients to attach to.
	        // we need an EventSource per TickerSubscriber because of assumptions
	        // make in StreamListener.
	        eventSource = wm.createEventSource();
            tickerUpdateProxy = (TickerUpdateEvents) eventSource.getEventTrigger(TickerUpdateEvents.class);

	        // We add a listener here so that we get listener count changes from the EventSource.
	        // once the listener count falls to 1, i.e. we're all thats left, then we release the thread the
	        // TickerSubscriber.
	        eventSource.addListener(this);

	        // we pass in true because this is a long running Work and hence needs
	        // to be handled differently than a pooled Work by the runtime.
	        me = wm.startWork(this, true);
    	}
    	finally
    	{
    		logger.exiting(className, methodName);
    	}
    }

    /**
     * This when called will cause the run method to exit. This is normally only called when the
     * JMS listener is no longer required.
     */
    public synchronized void release()
    {
        logger.logp(Level.FINEST, className, "release", "release called");
        released = true;
    }

	synchronized boolean getReleased()
	{
		return released;
	}

	/**
	 * This is the actual 'thread' which listens for prices.
	 * It allows the streaming servlet for an applet to register
	 * a listener that implements TickerUpdateEvents. We fire an event for each message received. All subscribers on the
	 * EventSource get this event. In this way, we share a single JMS listener for all subscribers within this
	 * JVM.
	 */
    public void run()
    {
        String methodName = "run";
        logger.entering(className, methodName);
    	TopicConnection jmsConn = null;
        try
        {
	    	TopicSession jmsSession = null;
	        // Connect to our Price Generator topic and subscribe to it
	        // using a filter for or specified ticker
	        InitialContext ic = new InitialContext();

	        // Get a JMS session for this Work. Note, that this is closed indirectly in the finally block at the
	        // end of this method. Closing the Connection also closes the session and subscriber.
	        jmsConn = tcf.createTopicConnection();
	        jmsSession = jmsConn.createTopicSession(false, TopicSession.AUTO_ACKNOWLEDGE);
            logger.logp(Level.FINE, className, methodName, "made session");

	        // Make a TopicSubscriber which will be used to receive the prices for the ticker.
	        // In this sample, all prices are published on the same topic and we use
	        // a selector to filter this to receive only the prices that we're interested in.
            TopicSubscriber subscriber = jmsSession.createSubscriber(topic, "ticker='" + ticker + "'", false);
            logger.logp(Level.FINE, className, methodName, "made subscriber");

			// need to start it to receive messages.
			jmsConn.start();

            // Continuously receive messages from the topic
            // and send them as notifications to the listeners until we're 'released'.
            while(!getReleased())
            {
		        // Wait for up to 5 seconds so that we don't block indefinitely. This gives us an opportunity
		        // to exit the run method if release is called.
                Message msg = subscriber.receive(5000);

                if (msg != null)
                {
                    logger.logp(Level.FINEST, className, methodName, "Received message", msg);

                    // Fire an event on our TickerUpdateEvents listeners
                    tickerUpdateProxy.tickerUpdated(this, msg);
                }
            }
            // this isn't used by the sample but illustrates the sorts of events
            // that can be fired by these types of objects.
            tickerUpdateProxy.tickerStopped(this);
        }
        catch (JMSException e)
        {
            logger.logp(Level.SEVERE, className, methodName, "Exception: " + e);
        }
        catch (NamingException e)
        {
            logger.logp(Level.SEVERE, className, methodName, "Exception: " + e);
        }
        catch (Throwable e)
        {
            logger.logp(Level.SEVERE, className, methodName, "Exception: " + e);
        }
        finally
        {
        	try
        	{
	        	if(jmsConn != null)
	        		jmsConn.close();
        	}
        	catch(JMSException e)
        	{
                logger.logp(Level.SEVERE, className, methodName, "Exception: " + e);
        	}
        	logger.exiting(className, methodName);
        }
    }

	/**
	 * This returns the EventSource associated with this listener. We fire an event when we
	 * receive a message. Normally, the only listeners are ones which store the message in a
	 * modified FIFO queue where another thread picks it up for processing at a later point.
	 */
	public EventSource getEventSource()
	{
		return eventSource;
	}

    /**
     * If the listener counter falls to one listener, then that
     * means we're the only listener and therefore we will stop the
     * Work by calling release and this will stop us listening needlessly when
     * no one is listening to the received messages.
     * @see com.ibm.websphere.asynchbeans.EventSourceEvents#listenerCountChanged(EventSource, int, int)
     */
    public void listenerCountChanged(
        EventSource es,
        int oldCount,
        int newCount)
    {
        String methodName = "listenerCountChanged";
        logger.entering(className, methodName);
        try
        {
        	// listeners can be registered with multiple eventsources, verify that this
        	// is being called by the right one. This listener is only registered with
        	// a single eventsource in this sample but this is intended to highlight
        	// that you need to do this if registering a listener with multiple
        	// event sources.
            if(es == eventSource)
            {
            	// when the count drops from 2 to 1 then the last subscriber has
            	// deregistered with the event source and there are now no interested
            	// parties.
                if(newCount == 1 && oldCount == 2)
                {
                    // we're the only listener so stop this listener.
                    release();
                    appState.removeTickerSubscriber(ticker);
                }
            }
        }
        finally
        {
            logger.exiting(className, methodName);
        }
    }

    /**
     * @see com.ibm.websphere.asynchbeans.EventSourceEvents#listenerExceptionThrown(EventSource, Object, String, Throwable)
     */
    public void listenerExceptionThrown(
        EventSource es,
        Object listener,
        String methodName,
        Throwable exception)
    {
    }

    /**
     * @see com.ibm.websphere.asynchbeans.EventSourceEvents#unexpectedException(EventSource, Object, Throwable)
     */
    public void unexpectedException(
        EventSource es,
        Object runnable,
        Throwable exception)
    {
    }

}
