//
// "This program may be used, executed, copied, modified and distributed without royalty for the
// purpose of developing, using, marketing, or distributing."
//
// Product 5724i63,  (C) COPYRIGHT International Business Machines Corp., 2001,2004
// All Rights Reserved * Licensed Materials - Property of IBM
// US Government Users Restricted Rights - Use, duplication or disclosure
// restricted by GSA ADP Schedule Contract with IBM Corp.
//
package com.ibm.websphere.samples.asynchbeans.wstrader.tickstream.web;

import java.io.IOException;
import java.io.PrintWriter;

import javax.servlet.ServletException;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import java.util.logging.*;
import com.ibm.websphere.samples.asynchbeans.wstrader.ClientInstruction;
import com.ibm.websphere.samples.asynchbeans.wstrader.Config;
import com.ibm.websphere.samples.asynchbeans.wstrader.tickstream.StreamClientState;
import com.ibm.websphere.samples.asynchbeans.wstrader.tickstream.StreamListener;
import com.ibm.websphere.samples.asynchbeans.wstrader.tickstream.TickerSubscriber;

/**
 * This is a simple implementation of HTTP push. A HTTP client sends a HTTP request to this servlet and
 * we then push events back asynchronously using the HTTP response. We actually block the servlet engine
 * thread that received the request to do this. The servlet thread enters a loop in the StreamListener.sendMessages
 * method and this loop waits for an asynchronous event. When one arrives then it writes the event to the response
 * and then flushes it. The applet will then get it. Next, the sendMessages method simply loops around and waits
 * for another event. When the client logs out or times out then sendMessages is signaled and returns to this
 * servlet which then returns to the servlet engine.<p>
 * Note, this <b>is</b> a sample. We are holding a thread per concurrent streamer using this implementation. This does not
 * scale very well. If you wanted to attach 1000s of clients to a server then you'd need 1000s of threads in the servlet
 * engine thread pool. There would be too much thrashing. There are better approaches than the simplistic one
 * shown here. For example, we may use a third party HTTP streaming product, the server job is then to push events
 * to this product when the occur and the product then pushes it to the client using HTTP. Lotus Sametime could be
 * used to do this sort of streaming as could other types of IM product.<p>
 * Another alternative would be JMS tunnelled over HTTP. You may think that then the applet can talk directly to the
 * JMS pub/sub to get the tickers but you'd probably be surprised. Imagine we had a high volume feed sending 10s of
 * prices per second per ticker. An applet attaching over a slow connection would not be able to keep up. You
 * could use non persistent JMS and then the broker would start dropping messages ad hoc. However, you may also want
 * more control and do your own throttling of messages. A more useful approach in this scenarion would be to
 * hav a JMS client in the applet, possibly tunnelled over HTTP. The client makes a temporary queue and sends this
 * to the server. The server then pushes the events to the client by sending them to the queue. The server can do
 * additional filtering/processing on the events to tune the flow of events to the bandwidth of the client.<p>
 * So, the point is this sample won't scale well. There are better approaches but the purpose here was simply
 * to demonstrate how HTTP push could be used.
 */
public class StreamingServlet extends BaseAsynchServlet
{
    static final Logger logger = Logger.getLogger(StreamingServlet.class.getName());
    static final String className = StreamingServlet.class.getName();

	/**
	 * Constructor for TickerStreamerServlet.
	 */
	public StreamingServlet()
    {
		super();
	}

    /**
     * @see javax.servlet.http.HttpServlet#doGet(HttpServletRequest, HttpServletResponse)
     */
    protected void doGet(HttpServletRequest req, HttpServletResponse resp)
        throws ServletException, IOException
    {
        doProcess(req, resp);
    }

    /**
     * @see javax.servlet.http.HttpServlet#doPost(HttpServletRequest, HttpServletResponse)
     */
    protected void doPost(HttpServletRequest req, HttpServletResponse resp)
        throws ServletException, IOException
    {
        doProcess(req,resp);
    }

	boolean isStringSpecified(String s)
	{
		return (s != null && s.trim().length() > 0);
	}

    /**
     * Handle all client requests.
     * @param req the client request
     * @param resp the response to the client
     */
    protected void doProcess(HttpServletRequest req, HttpServletResponse resp)
        throws ServletException, IOException
    {
        String methodName = "doProcess";
        logger.entering(className, methodName);

        PrintWriter pw = resp.getWriter();
        TickerSubscriber curTickSub = null;
        StreamListener servletListener = null;
        try
        {
        	setTextResponse(resp, true);

	        // Grab the client id from the cookie.
            String loginid = getCookieValue(req.getCookies(), Config.CLIENT_COOKIE_NAME, Config.CLIENT_PARAM_CLIENTID);
            String streamName = req.getParameter(Config.CLIENT_PARAM_STREAM_NAME);
            String tickers = req.getParameter(Config.CLIENT_PARAM_TICKERS);

            logger.logp(Level.FINEST, className, methodName, "Parameters:", new Object[]{loginid,streamName,tickers});

	        // If no cookie or client id, then send back an error response.
	        if (!isStringSpecified(loginid))
	        {
                loginid = req.getParameter(Config.CLIENT_PARAM_CLIENTID);
                if (!isStringSpecified(loginid))
                {
                    logger.logp(Level.FINE, className, methodName, "No valid client id specified in cookie or querystring");
	                pw.print(ClientInstruction.createErrorResponse(ClientInstruction.RESPONSE_ERRORMSG_NOLOGIN));
                    pw.flush();
	                return;
                }
	        }

            // Get the state of our client
			StreamClientState clientState = getClientState(loginid);

            if (clientState == null)
            {
                logger.logp(Level.FINE, className, methodName, "Client State is null");
                pw.print(ClientInstruction.createErrorResponse(ClientInstruction.RESPONSE_ERRORMSG_NOLOGIN));
                pw.flush();
                return;
            }
            logger.logp(Level.FINE, className, methodName, "Found tickers in state " + clientState.getClientID() + ": " + tickers);

            // Lock this thread to keep the client response socket open.
            // The Servlet Listener will write information to our stream for us from
            // this point on.
            pw.println("Processing...");
            pw.flush();

            // Create our Listener for this client and streamer.
            servletListener = new StreamListener(clientState, pw, streamName);

            // Update our Listener for this client with the updated tickers.
            servletListener.updateTickers(tickers, false);

            // Start actively pushing the messages to the client.
            // This method will block until the client ends.
			servletListener.sendMessages();
            logger.logp(Level.FINE, className, methodName, "sendMessages is complete.  StreamingServlet ending.");

        }
        catch (Exception e)
        {
            String message = "StreamingServlet() threw Exception";
            pw.print(ClientInstruction.createErrorResponse(message, e));
            logger.logp(Level.SEVERE, className, methodName, "Exception processing client requests: " + e);
        }
        finally
        {
            logger.exiting(className, methodName);
        }
    }
}
