//
// "This program may be used, executed, copied, modified and distributed without royalty for the 
// purpose of developing, using, marketing, or distributing."
//
// Product 5724i63,  (C) COPYRIGHT International Business Machines Corp., 2001, 2004
// All Rights Reserved * Licensed Materials - Property of IBM
//
 
package com.ibm.websphere.samples.jtaextensionssamples.transactiontrackerwar;

import java.io.PrintWriter;
import java.util.Vector;

import javax.naming.InitialContext;
import javax.naming.NamingException;
import javax.servlet.http.HttpServlet;
import javax.transaction.UserTransaction;

import com.ibm.websphere.jtaextensions.ExtendedJTATransaction;
import com.ibm.websphere.jtaextensions.NotSupportedException;
import com.ibm.websphere.jtaextensions.SynchronizationCallback;

/** The TransactionTracker servlet provides a demonstration of WebSphere's
 *  Extended JTA Transaction functionality.
 */
public class TransactionTracker extends HttpServlet implements SynchronizationCallback
{
    private static ExtendedJTATransaction extJTATran = null;
    private static UserTransaction ut = null;

    // Stores details of each completion call.
    private static Vector completionCalls = new Vector();

    /** Initialize the TransactionTracker servlet. This method
     *  is driven when the servlet is first loaded and is
     *  used to register the synchronization callback with
     *  ExtendedJTATransaction and lookup UserTransaction.
     */
    public void init() throws javax.servlet.ServletException
    {
        // Lookup ExtendedJTATransaction and UserTransaction
        try
        {
            InitialContext ctx = new InitialContext();
            extJTATran = (ExtendedJTATransaction) ctx.lookup("java:comp/websphere/ExtendedJTATransaction");
            ut = (UserTransaction) ctx.lookup("java:comp/UserTransaction");
        }
        catch (NamingException ne)
        {
            System.out.println("Error looking up object: " + ne);
        }

        // If the lookup succeeded, register a callback.
        if (extJTATran != null)
        {
            try
            {
                // Register this class as a SynchronizationCallback.
                // This class will now be notified of all transactions 
                // that undergo completion.
                extJTATran.registerSynchronizationCallback(this);
            }
            catch (NotSupportedException nse)
            {
            }
        }
    }

    /** Act upon a post request to the TransactionTracker.
     *  This method processes requests to run a new transaction
     *  that is either committed or rolledback. It outputs the
     *  result of the attempt to run the transaction as well as
     *  a table detailing all of the transactions that have
     *  undergone completion following the callback being
     *  registered at servlet initialization time.
     */
    public void doPost(javax.servlet.http.HttpServletRequest request, javax.servlet.http.HttpServletResponse response) throws javax.servlet.ServletException, java.io.IOException
    {
        response.setContentType("text/html");

        PrintWriter out = response.getWriter();
        outputHeader(out);

        String txResultString = "";

        // We can only run a transaction if the lookup of
        // UserTransaction at initialization time was
        // successful.
        if (ut != null)
        {
            try
            {
                // Start a new transaction.
                ut.begin();

                // Find the value of the trantype parameter that
                // dictates whether the transaction should be
                // committed or rolledback.
                final String trantype = (String)request.getParameter("trantype");

                // The request is for the transaction
                // to be committed.
                if ((trantype != null) && (trantype.equals("commit")))
                {
                    ut.commit();
                    txResultString = "Transaction committed";
                }
                // The request is for the transaction
                // to be rolledback.
                else
                {
                    ut.rollback();
                    txResultString = "Transaction rolledback";
                }
            }
            catch (Exception e)
            {
                // An error has occured while working with
                // the transaction - inform the user.
                out.println("<p>An error was encountered attempting to run the transaction:</p>");
                out.println("<p>" + e + "</p>");
            }

        }
        else
        {
            // The lookup of user transaction has failed - inform the user.
            txResultString = "User Transaction not found - could not run transaction";
        }

        // Output the form that allows the
        // user to run a transaction.
        outputRunTransaction(out, txResultString);

        // Output the table detailing all the
        // callbacks that have been received
        // since registration.
        outputCallbackTable(out);

        outputFooter(out);

        out.flush();
        out.close();

    }

    // Output a form that allows the user to
    // request that a transaction is run. The
    // request is made by making a post request
    // to this servlet with a trantype of
    // commit or rollback.
    private void outputRunTransaction(PrintWriter out, String txResultString)
    {
        out.println("<p><b>Transaction type:</b></p>");
        out.println("<form method=\"post\" action=\"TransactionTracker\">");
        out.println("<p><label for=\"commit\">Commit: <input type=\"radio\" name=\"trantype\" id=\"commit\" value=\"commit\" checked></label></p>");
        out.println("<p><label for=\"rollback\">Rollback: <input type=\"radio\" name=\"trantype\" id=\"rollback\" value=\"rollback\"></label></p>");
        out.println("<p><input type=\"submit\" value=\"Run Transaction\">&nbsp;&nbsp;&nbsp;" + txResultString + "</p>");
        out.println("</form>");
    }

    // Output a table of all the callbacks that have
    // been received since registration or if no
    // callbacks have been received report this to
    // the user.
    private void outputCallbackTable(PrintWriter out)
    {
        // Output the table header
        out.println("<table border=\"1\">");
        out.println("<tr>");
        out.println("<th valign=\"top\">Method Called</th><th valign=\"top\">Committed</th><th valign=\"top\">Local ID</th><th valign=\"top\">Global ID (94 characters)</th>");
        out.println("</tr>");

        // Transactions have begun completion since the synchronization
        // was registered so output their details in the table.
        if (completionCalls.size() > 0)
        {
            // Clone the Vector incase transactions complete
            // while the table is being generated.
            Vector callsClone = (Vector) completionCalls.clone();

            // For every call to one of the completion methods
            // recorded in the Vector output a row in the table.
            // Each row details the type of the call (either
            // beforeCompletion or afterCompletion), the local id,
            // the global id, and if applicable whether or not
            // the transaction was committed successfully.
            for (int i = 0; i < callsClone.size(); i++)
            {
                CompletionCall cc = (CompletionCall)callsClone.elementAt(i);

                out.println("<tr>");

                String localId = "" + cc.getLocalId();
                String globalId = byteArrayToHex(cc.getGlobalId());
                String committed = (cc.getType() == CompletionCall.BEFORE) ? "n/a" : "" + cc.getCommitted();
                String type = (cc.getType() == CompletionCall.BEFORE) ? "beforeCompletion" : "afterCompletion";

                out.println("<td>" + type + "</td><td>" + committed + "</td><td>" + localId + "</td><td>" + globalId + "</td>");

                out.println("</tr>");
            }
        }
        else
        {
            out.println("<tr><td colspan=\"4\">No transactions have completed since callback registration</td></tr>");
        }

        out.println("</table>");

        out.println("<form action=\"TransactionTracker\" method=\"GET\">");
        out.println("<input type=\"submit\" value=\"Refresh table\">");
        out.println("</form>");
    }

    // Output the start of the HTML response.
    private void outputHeader(PrintWriter out)
    {
        out.println("<html>");
        out.println("<script language=\"JavaScript\" src=\"/WSsamples/scripts/browser_detection.js\"></script>");

        out.println("<NOSCRIPT>");
        out.println("Proper display of samples requires JavaScript to be enabled.");
        out.println("</NOSCRIPT>");

        out.println("<HEAD>");
        out.println("<TITLE>JTA Extensions - Transaction Tracker</TITLE>");
        out.println("</HEAD>");

        out.println("<BODY leftmargin=0 topmargin=0 marginwidth=\"0\" marginheight=\"0\">");

        out.println("<img src=\"/WSsamples/images/success_banner.jpg\" valign=top alt=\"WebSphere Samples\">");

        out.println("<table cellpadding=5 cellspacing=5>");
        out.println("<tr>");
        out.println("<td>");
        out.println("<BR>");
        out.println("<H1 class=\"sampjsp\">JTA Extensions - Transaction Tracker</H1>");
        out.println("</td>");
        out.println("</tr>");
        out.println("</table>");

        out.println("<HR width=100%>"); 

        out.println("<table cellpadding=5 cellspacing=5>");
        out.println("<tr>");
        out.println("<td>");
    }

    // Output the end of the HTML response
    private void outputFooter(PrintWriter out)
    {
        out.println("</td>");
        out.println("</tr>");
        out.println("</table>");

        out.println("</body>");
        out.println("</html>");
    }

    /** Process a get request to the TransactionTracker.
     *  Output a form allowing the user to request that
     *  a transaction is run as well as a table of all
     *  callbacks that have been received since registration.
     */
    public void doGet(javax.servlet.http.HttpServletRequest request, javax.servlet.http.HttpServletResponse response) throws javax.servlet.ServletException, java.io.IOException
    {
        response.setContentType("text/html");
        
        PrintWriter out = response.getWriter();

        outputHeader(out); 
        outputRunTransaction(out, "");
        outputCallbackTable(out);
        outputFooter(out);

        out.flush();
        out.close();
    }

    // Convert the given byte[] to a hex String
    private String byteArrayToHex(byte[] array)
    {
        int dumpLength = array.length;

        StringBuffer hex = new StringBuffer();

        for (int i = 0; i < array.length; i++)
        {
            int binByte = (int)array[i];

            // Treat the byte as unsigned
            if ( binByte<0 )  binByte += 256;

            // Ensure we generate 2 chars per byte
            if ( binByte<16 ) hex.append('0');

            hex.append( Integer.toHexString(binByte).toUpperCase() );
        }

        return(hex.toString());
    }

    /** This method is invoked before the completion of a transaction.
      * For the purpose of this sample we
      * simply store the details of the transaction involved in
      * the Vector which is used to generate the table of
      * completion calls.
      */
    public void beforeCompletion(int localId, byte[] globalId)
    {       
        // Record the invocation of beforeCompletion in the vector.
        completionCalls.addElement(new CompletionCall(localId, globalId));          
    }

    /** This method is invoked after the completion of a transaction.
      * For the purpose of this sample we
      * simply store the details of the transaction involved in
      * the Vector which is used to generate the table of
      * completion calls.
      */
    public void afterCompletion(int localId, byte[] globalId, boolean committed)
    {
        // Record the invocation of afterCompletion in the vector.
        completionCalls.addElement(new CompletionCall(localId, globalId, committed));               
    }

    // Inner-class that encapsulates the data passed
    // to either of the completion methods defined
    // by the SynchronizationCallback interface.
    private class CompletionCall
    {
        // Indicates that this was a call to beforeCompletion;
        public static final int BEFORE = 0;

        // Indicates that this was a call to afterCompletion;
        public static final int AFTER = 1;

        // The localId of the transaction.
        private int localId;

        // The globalId of the transaction.
        private byte[] globalId;

        // Records whether or not the transaction
        // was committed successfully.
        boolean committed;

        int type;

        // Create a new CompletionCall for the beforeCompletion case.
        public CompletionCall(int localId, byte[] globalId)
        {
            this.localId = localId;
            this.globalId = globalId;
            type = BEFORE;
        }

        // Create a new CompletionCall for the afterCompletion case.
        public CompletionCall(int localId, byte[] globalId, boolean committed)
        {
            this.localId = localId;
            this.globalId = globalId;
            type = AFTER;
            this.committed = committed; 
        }

        // Return the local id of the transaction
        public int getLocalId()
        {
            return localId;
        }

        // Return the global id of the transaction
        public byte[] getGlobalId()
        {
            return globalId;
        }

        // Return the type of the completion call
        // that this object encapsulates
        public int getType()
        {
            return type;
        }

        // Return whether or not the transaction's
        // completion resulted in it being committed.
        public boolean getCommitted()
        {
            return committed;
        }
    } 
}

