//
// "This program may be used, executed, copied, modified and distributed without royalty for the 
// purpose of developing, using, marketing, or distributing."
//
// Product 5724i63,  (C) COPYRIGHT International Business Machines Corp., 2001,2004
// All Rights Reserved * Licensed Materials - Property of IBM
// US Government Users Restricted Rights - Use, duplication or disclosure
// restricted by GSA ADP Schedule Contract with IBM Corp.
//
package com.ibm.websphere.samples.scheduler.accountreportejb;

import java.util.Date;

import javax.ejb.CreateException;
import javax.ejb.SessionBean;
import javax.ejb.SessionContext;

import com.ibm.websphere.scheduler.UserCalendarPeriodInvalid;
import com.ibm.websphere.scheduler.UserCalendarSpecifierInvalid;

/**
 * Implements a UserCalendar which will increment a given date
 * by N seconds.
 * 
 * UserCalendar session beans can implement the UserCalendarBean
 * interface so that the method signatures are checked at compile
 * time.
 */
public class SampleCalendarBean implements SessionBean {

    /**
     * Checks that a given delta String is suitable for use with this calendar.
     * 
     * @param calendar The calendar to check
     * @param delta    The delta to be checked
     * @exception UserCalendarSpecifierInvalid
     *                   if the calendar parameter is not "seconds"
     * @exception UserCalendarPeriodInvalid
     *                   if the delta is not suitable for use with this calendar
     */
    public void validate(String calendar, String delta) throws UserCalendarSpecifierInvalid, UserCalendarPeriodInvalid{        
        //to check the delta specified, just call applyDelta with the current time
        Date currentTime = new Date(System.currentTimeMillis());
        applyDelta(currentTime, calendar, delta);
    }

    /**
     * Increments the base time by a delta. The only supported
     * 
     * @param baseTime The base time to which the increment will be added
     * @param calendar Used by this calendar bean to specify the unit to be scaled.
     *                 
     *                 The only accepted value at this time is "seconds"
     * @param delta    A scalar which determines how much to add onto the base time. The
     *                 unit to be scaled is determined by the calendar parameter.
     * @return The new date which is the baseTime + delta
     * @exception UserCalendarSpecifierInvalid
     *                   if the calendar parameter is not "seconds"
     * @exception UserCalendarPeriodInvalid
     *                   if the delta is not suitable for use with this calendar
     */
    public java.util.Date applyDelta(java.util.Date baseTime, String calendar, String delta) throws UserCalendarSpecifierInvalid, UserCalendarPeriodInvalid{
        //check to see if the calendar specified is seconds
        if (calendar.equals("seconds")){
            try {
                //the delta should be an integer, but must be converted from it's String representation
                //when the time is incremented, the millisecond value of this number will be used, so calculate this as well
                long incrementMillis = Long.parseLong(delta) * 1000;
                
                if(incrementMillis < 0)
                    throw new UserCalendarPeriodInvalid("Only positive whole numbers are supported.");
    
                //now calculate the new time by adding the base time to the increment value
                Date returnValue = new Date(baseTime.getTime() + incrementMillis);
                
                return(returnValue);
            }
            catch(NumberFormatException e) {
                throw new UserCalendarPeriodInvalid("Only positive whole numbers are supported.");
            }
        } else{
            throw new UserCalendarSpecifierInvalid("\"seconds\" is the only supported calendar specifier");
        }
    }

    /**
     * Retrieves an array of valid calendar names that this UserCalendar implementation supports.
     * @return the array of calendar names.
     */    
    public String[] getCalendarNames() 
    {
        return new String[]{"seconds"};
    }

    private SessionContext ejbContext;


    /**
     * A container invokes this method when it creates the EJB object     
     *
     * @exception EJBException Thrown by the method to indicate a failure caused by a system-level error.
     */
    public void ejbCreate() throws CreateException
    {
    }


    /**
     * A container invokes this method when the instance is
     * taken out of the pool of available instances to become
     * associated with a specific EJB object.
     *
     * @exception EJBException Thrown by the method to indicate a failure caused by a system-level error.
     */
    public void ejbActivate()
    {
    }

    /**
     * A container invokes this method on an instance before
     * the instance becomes disassociated with a specific EJB object.
     *
     * @exception EJBException Thrown by the method to indicate a failure caused by a system-level error.
     */
    public void ejbPassivate()
    {
    }

    /**
     * A container invokes this method before it removes the EJB object
     * that is currently associated with the instance.
     *
     * @exception EJBException Thrown by the method to indicate a failure caused by a system-level error.
     */
    public void ejbRemove()
    {
    }

    /**
     * Set the associated session context.
     *
     * @param ctx A SessionContext interface for the instance. The instance should store the reference
     *            to the context in an instance variable.
     * @exception EJBException Thrown by the method to indicate a failure caused by a system-level error.
     */
    public void setSessionContext(SessionContext ctx)
    {
        ejbContext = ctx;
    }
}
