/* "This sample program is provided AS IS and may be used, executed, copied and modified without royalty payment by customer (a) for its own 
 * instruction and study, (b) in order to develop applications designed to run with an IBM WebSphere product, either for customer's own internal use 
 * or for redistribution by customer, as part of such an application, in customer's own products. " 
*/
 
package com.ibm.websphere.samples.technologysamples.sdo.AlbumCatalog;

import java.sql.Connection;
import java.sql.SQLException;
import java.util.HashMap;
import java.util.Map;

import javax.naming.InitialContext;
import javax.sql.DataSource;

import com.ibm.websphere.sdo.mediator.JDBCMediator;
import com.ibm.websphere.sdo.mediator.exception.MediatorException;
import com.ibm.websphere.sdo.mediator.jdbc.ConnectionWrapper;
import com.ibm.websphere.sdo.mediator.jdbc.JDBCMediatorFactory;
import com.ibm.websphere.sdo.mediator.jdbc.exception.InvalidMetadataException;
import com.ibm.websphere.sdo.mediator.jdbc.metadata.Metadata;
import com.ibm.websphere.sdo.mediator.jdbc.ConnectionWrapperFactory;
import commonj.sdo.DataObject;

/**
 * AlbumCatalogJDBCSDOService provides methods to retrieve and commit Datagraphs from and to the Album Catalog relational database.
 * It  uses AlbumCatalogMetadata and JDBC Mediator APIs to retrieve/commit SDOs.
 *  
 */

public class AlbumCatalogJDBCSDOService {
		 
	private static AlbumCatalogJDBCSDOService soleInstance;
	
	// JNDI name of Data Source 
	private static String ALBUM_DS_JNDI = "java:comp/env/jdbc/ALBUM";		
	
	// cache DataSource 
	private static DataSource albumDS;
	
	public static AlbumCatalogJDBCSDOService soleInstance() 
	{
		
		if(soleInstance == null)
		{			
			initDataSource();
			soleInstance = new AlbumCatalogJDBCSDOService();
		}
		
		return soleInstance;
	}
	
	private static void initDataSource()
	{
		try
		{
			InitialContext initialCtx = new InitialContext();
			albumDS = (DataSource)initialCtx.lookup(ALBUM_DS_JNDI);			
		}
		catch(Exception e)
		{
			e.printStackTrace();		
		}

	}

	/**
	 * Uses JDBC Mediator APIs to find bands using the parameter passed in and return the results as a Datagraph
	 * Note that this method uses JDBCMediator Metadata APIs (from the AlbumCatalogMetadata class) to describe
	 * the query.         
	 *  
	 * @param  bandToFind	the name of band to search.
	 * @return DataObject 	Datagraph of Band DataObjects       	 
	 */	
	public DataObject findBand(String bandToFind) throws MediatorException
	{											
		// specify if we want the band DataObjects to be the top level DataObjects in the graph 		
		boolean isBandRoot = true;		

		// a non-exclusive relationship in relational metadata is similar to a left outer join; 
		// i.e. bands which have no related albums in the album table, will be included in the Datagraph. 
		boolean isAlbumRelationshipExclusive = false;		
		
		// get relational metadata for the Band table and Album table, with a non-exclusive relationship between band and album tables  
		Metadata metadata = AlbumCatalogMetadata.getBandAndAlbumMetadata(isBandRoot, isAlbumRelationshipExclusive);		
		
		// add a filter for the band name field of the band table. Please note that the jdbc mediator provides two options for describing
		// the SQL query: 1) Through JDBC Mediator Metadata APIs, as shown in the case of findBand operation, 2) Through a user-supplied SQL string,
		// as shown in the findAlbum operation.
		AlbumCatalogMetadata.addBandNameFilter(metadata);
				
		// get connection to the database
		ConnectionWrapper albumConnWrapper = getAlbumConnectionWrapper();

		// create mediator with metadata and connection information. 
		JDBCMediator albumDBMediator = createMediator(metadata, albumConnWrapper);
		
		// when metadata has filters - in this case a filter for the band name field - parameters for the filters are passed
		// as DataObjects 								
		try
		{
			 
			DataObject parameterDataObject = albumDBMediator.getParameterDataObject();
			
			// note that "bandName" is the name of the filter defined in the metadata.
			parameterDataObject.setString("bandName", bandToFind);
			
			// parameter data object is passed in to the getGraph method.
			DataObject bandGraph = albumDBMediator.getGraph(parameterDataObject);
			
			return bandGraph;
		}
		catch(MediatorException me)			
		{
			throw me;
		}
		finally
		{
			closeConnection(albumConnWrapper);
		}				
		
	}

	
	/**
	 * Uses JDBC Mediator APIs to find an album using the parameter passed in and return the results as a Datagraph.
	 * Note that this method uses a SQL query string, instead of using Metadata APIs, to describe the query	          
	 *  
	 * @param  albumToFind	the name of album to search.
	 * @return DataObject	Datagraph of album DataObjects       	 
	 */	
	public DataObject findAlbum(String albumToFind) throws MediatorException
	{
		
		ConnectionWrapper albumConnWrapper = getAlbumConnectionWrapper();

		// get relational metadata for the Album table only; all columns used in the select clause of the SQL query need to be defined
		// in the metadata
		Metadata metadata = AlbumCatalogMetadata.getAlbumMetadata();
		
		//construct sql query
		StringBuffer sqlQuery = new StringBuffer("select * from ");		
		sqlQuery.append(AlbumCatalogMetadata.ALBUMTABLE); 
		sqlQuery.append(" where ");
		sqlQuery.append(AlbumCatalogMetadata.ALBUM_NAME_FIELD);
		sqlQuery.append(" LIKE '");
		sqlQuery.append(albumToFind);
		sqlQuery.append("'");
		
		// create mediator with metadata, sql string and connection
		JDBCMediator albumDBMediator = createMediator(metadata, sqlQuery.toString(), albumConnWrapper);		

		try
		{
			 
			DataObject albumGraph = albumDBMediator.getGraph();		
		
			return albumGraph;
		}
		catch(MediatorException me)			
		{
			throw me;
		}
		finally
		{
			closeConnection(albumConnWrapper);
		}				

	}
	
	/**
	 * Uses JDBC Mediator APIs to find albums by band name and year passed in and returns the results as a Datagraph.        
	 *  
	 * @param  bandName	the name of band to search.
	 * @param  albumYear	the album year to search.
	 * @return DataObject	Datagraph of album DataObjects       	 
	 */	
	public DataObject findAlbumsByBandAndYear(String bandName, String albumYear) throws MediatorException
	{
				
		ConnectionWrapper albumConnWrapper = getAlbumConnectionWrapper();
		Metadata metadata = AlbumCatalogMetadata.getAlbumMetadata();
		
		//construct sql query. Note that only columns used in the select clause need to be in the metadata
		StringBuffer sqlQuery = new StringBuffer("SELECT ");
		sqlQuery.append(AlbumCatalogMetadata.ALBUM_BANDNAME_FIELD +",");
		sqlQuery.append(AlbumCatalogMetadata.ALBUM_NAME_FIELD + ",");
		sqlQuery.append(AlbumCatalogMetadata.ALBUM_YEAR_FIELD + ",") ;
		sqlQuery.append(AlbumCatalogMetadata.ALBUM_TRACKS_FIELD);		
		sqlQuery.append(" FROM ");
		sqlQuery.append(AlbumCatalogMetadata.ALBUMTABLE);
		sqlQuery.append(" INNER JOIN "); 
		sqlQuery.append(AlbumCatalogMetadata.BANDTABLE);
		sqlQuery.append(" ON ");
		sqlQuery.append(AlbumCatalogMetadata.BAND_NAME_FIELD  + " LIKE '" + bandName + "'"); 
		sqlQuery.append(" AND " + AlbumCatalogMetadata.ALBUM_YEAR_FIELD  + " = " + albumYear);
		sqlQuery.append(" AND " + AlbumCatalogMetadata.BAND_NAME_FIELD  + " = " + AlbumCatalogMetadata.ALBUM_BANDNAME_FIELD);
				
		JDBCMediator albumDBMediator = createMediator(metadata, sqlQuery.toString(), albumConnWrapper);
												
		try
		{
			DataObject albumGraph = albumDBMediator.getGraph();
		
			return albumGraph;
		}
		catch(MediatorException me)			
		{
			throw me;
		}
		finally
		{
			closeConnection(albumConnWrapper);
		}				

	}

	
	/**
	 * Uses JDBC Mediator APIs create an empty Datagraph with relational metadata for the Band table. Therefore only
	 * DataObjects that contain fields for the Band table can be added to this Datagraph.            
	 *  
	 * @return DataObject	empty Datagraph for Band metadata       	 
	 */	
	public DataObject getEmptyBandGraph() throws MediatorException
	{		
		// note that a connection is not needed to create an empty Datagraph
		return getEmptyGraph(AlbumCatalogMetadata.getBandMetadata());						
		
	}

	/**
	 * Uses JDBC Mediator APIs to create an empty Datagraph with relational metadata for the Album table. Therefore only
	 * DataObjects that contain fields for the Album table can be added to this Datagraph.            
	 *  
	 * @return DataObject empty Datagraph for Album metadata       	 
	 */	
	public DataObject getEmptyAlbumGraph() throws MediatorException
	{
		
		//	note that a connection is not needed to create an empty Datagraph
		return getEmptyGraph(AlbumCatalogMetadata.getAlbumMetadata());
		
	}

	/**
	 * Uses JDBC Mediator APIs to create an empty Datagraph for relational metadata passed in.        
	 *  
	 * @return graph empty Datagraph for Album metadata       	 
	 */	
	
	public DataObject getEmptyGraph(Metadata metadata) throws MediatorException
	{
		
		//	note that a connection is not needed to create an empty Datagraph
		JDBCMediator albumDBMediator = createMediator(metadata);
		
		return albumDBMediator.getEmptyGraph();
		
	}

	/**
	 * Commits Datagraph containing album DataObjects to the database.     	             
	 *  
	 * @param graph the Datagraph to commit       	 
	 */	
	public void commitAlbumGraph(DataObject graph) throws MediatorException
	{
		
		ConnectionWrapper albumConnWrapper = getAlbumConnectionWrapper();				    				
		
		// create mediator with metadata pertinent to the data in the graph; in this case, metadata for the album table  
		JDBCMediator albumDBMediator = createMediator(AlbumCatalogMetadata.getAlbumMetadata(), albumConnWrapper);
		
		// when auto-commit property of java.sql.Connection is set to false, the jdbc mediator handles transactions. 
		// if the application wants to manage transactions, then a "passive" ConnectionWrapper is needed, which can be obtained from the ConnectionWrapperFactory 		
		setAutoCommitFalse(albumConnWrapper);		
		
		try
		{
			// the applychanges method commits a Datagraph to the data source. Note that only modified and new data is committed to the db.
			albumDBMediator.applyChanges(graph);
		}
		catch(MediatorException me)			
		{
			throw me;
		}
		finally
		{
			closeConnection(albumConnWrapper);
		}
																													
	}

	/**
	 * Commits Datagraph containing Band and Album dataobjects the database.    	             
	 *  
	 * @param graph the Datagraph to commit       	 
	 */	
	public void commitBandWithAlbumsGraph(DataObject graph) throws MediatorException
	{		
	
		ConnectionWrapper albumConnWrapper = getAlbumConnectionWrapper();				 
	
		JDBCMediator albumDBMediator = createMediator(AlbumCatalogMetadata.getBandAndAlbumMetadata(true, false), albumConnWrapper);
								
		// when auto-commit property of java.sql.Connection is set to false, the jdbc mediator handles transactions. 
		// if the application wants to manage transactions, then a "passive" ConnectionWrapper is needed, which can be obtained from the ConnectionWrapperFactory 		
		setAutoCommitFalse(albumConnWrapper);
	
		try
		{
			// the applychanges method commits a Datagraph to the data source. Note that only modified and new data is committed to the db.
			albumDBMediator.applyChanges(graph);
		}
		catch(MediatorException me)
		{			
			throw me;				
		}
		finally
		{
			closeConnection(albumConnWrapper);
		}
		
	}

	/**
	 * Commits Datagraph containing band DataObject to the database.    	             
	 *  
	 * @param graph the Datagraph to commit       	 
	 */	
	public void commitBandGraph(DataObject graph) throws MediatorException
	{
		
		ConnectionWrapper albumConnWrapper = getAlbumConnectionWrapper();
		
		JDBCMediator albumDBMediator = createMediator(AlbumCatalogMetadata.getBandMetadata(), albumConnWrapper);
								
		// when auto-commit property of java.sql.Connection is set to false, the jdbc mediator handles transactions. 
		// if the application wants to manage transactions, then a "passive" ConnectionWrapper is needed, which can be obtained from the ConnectionWrapperFactory 		
		setAutoCommitFalse(albumConnWrapper);
		
		try
		{
			// the applychanges method commits a Datagraph to the data source. Note that only modified and new data is committed to the db.
			albumDBMediator.applyChanges(graph);
		}
		catch(MediatorException me)
		{			
			throw me;				
		}
		finally
		{
			closeConnection(albumConnWrapper);
		}

	}

	private void setAutoCommitFalse(ConnectionWrapper albumConnWrapper)
	{
		
		try
		{
			albumConnWrapper.getConnection().setAutoCommit(false);
		}
		catch(SQLException e)
		{
			e.printStackTrace();
		}
	}


	/**
	 * Creates a JDBC Mediator for the relational Metadata passed in. This can be used to create empty Datagraphs;
	 * DataObjects added to these empty Datagraphs can only contain fields described in the metadata that was used
	 * to create the mediator.  
	 * 
	 * @param metadata	relational metadata
	 * @return Mediator a jdbc mediator 
	 */
	private JDBCMediator createMediator(Metadata metadata)
	{
		
		// the jdbc mediator factory
		JDBCMediatorFactory mFactory = JDBCMediatorFactory.soleInstance;
		
		JDBCMediator mediator = null;
		try
		{
			mediator = mFactory.createMediator(metadata);	
		}
		catch(InvalidMetadataException e)
		{
			e.printStackTrace();
		}
		
		return mediator;
	}

	
	/**
	 * Creates a JDBC Mediator for the relational Metadata passed in. This can be used to retrive and commit Datagraphs from/to
	 * the database specified by the ConnectionWrapper.
	 * Datagraphs retrieved/committed using this mediator can only contain fields described in the metadata that was used
	 * to create the mediator.  
	 * 
	 * @param metadata	relational metadata
	 * @param connectionWrapper connection to the database
	 * @return Mediator a jdbc mediator 
	 */
	private JDBCMediator createMediator(Metadata metadata, ConnectionWrapper connectionWrapper)
	{
		
		JDBCMediatorFactory mFactory = JDBCMediatorFactory.soleInstance;
		
		JDBCMediator mediator = null;
		try
		{
			
			mediator = mFactory.createMediator(metadata, connectionWrapper);	
		}
		catch(InvalidMetadataException e)
		{
			e.printStackTrace();
		}
		
		return mediator;
	}

	/**
	 * Creates a JDBC Mediator for the relational Metadata passed in.
	 * Upon calling the getGraph method on this mediator, the SQL query passed in is executed and the 
	 * results are returned as a Datagraph. 
	 * Datagraphs retrieved/committed using this mediator can only contain fields described in the metadata that was used
	 * to create the mediator.  
	 * 
	 * @param metadata	relational metadata
	 * @param suppliedSQL	SQL query to execute
	 * @param connectionWrapper connection to the database
	 * @return Mediator a jdbc mediator 
	 */
	private JDBCMediator createMediator(Metadata metadata, String suppliedSQL, ConnectionWrapper connectionWrapper)
	{
		
		JDBCMediatorFactory mFactory = JDBCMediatorFactory.soleInstance;
		
		JDBCMediator mediator = null;
		try
		{
			
			mediator = mFactory.createMediator(metadata, suppliedSQL, connectionWrapper);	
		}
		catch(InvalidMetadataException e)
		{
			e.printStackTrace();
		}
		
		return mediator;
	}

	
	/**
	 * gets a connection from the data source, uses it to instantiate a ConnectionWrapper object and returns
	 * the ConnectionWrapper.  
	 *  
	 * @return ConnectionWrapper
	 */
	private ConnectionWrapper getAlbumConnectionWrapper()
	{		 		 		
		
		Connection conn = null;
				
		try
		{
			conn = albumDS.getConnection();		
		}
		catch(SQLException se)
		{
			se.printStackTrace();
		}

		ConnectionWrapper connWrapper = ConnectionWrapperFactory.soleInstance.createConnectionWrapper(conn);
				
		return connWrapper;
		
	}

	/**
	 * closes connection passed in
	 */
	private void closeConnection(ConnectionWrapper connWrapper)
	{
		try
		{
			connWrapper.getConnection().close();
		}
		catch(SQLException e)
		{
			e.printStackTrace();
		}
	}

}
