/* "This sample program is provided AS IS and may be used, executed, copied and modified without royalty payment by customer (a) for its own 
 * instruction and study, (b) in order to develop applications designed to run with an IBM WebSphere product, either for customer's own internal use 
 * or for redistribution by customer, as part of such an application, in customer's own products. " 
*/

package com.ibm.websphere.samples.technologysamples.sdo.AlbumCatalog;

import com.ibm.websphere.sdo.mediator.jdbc.metadata.Column;
import com.ibm.websphere.sdo.mediator.jdbc.metadata.Filter;
import com.ibm.websphere.sdo.mediator.jdbc.metadata.FilterArgument;
import com.ibm.websphere.sdo.mediator.jdbc.metadata.Key;
import com.ibm.websphere.sdo.mediator.jdbc.metadata.Metadata;
import com.ibm.websphere.sdo.mediator.jdbc.metadata.MetadataFactory;
import com.ibm.websphere.sdo.mediator.jdbc.metadata.Relationship;
import com.ibm.websphere.sdo.mediator.jdbc.metadata.Table;

/**
 * AlbumCatalogMetadata provides relational metadata for the Album Catalog database.
 * 
 * <UL>
 * 	<LI> relational metadata is needed by the jdbc mediator to performs queries on the database. </LI>
 * 	<LI> relational metadata describes tables, columns, column types, primary keys, filters, and relationships to be used in a query.</LI>
 * 	<LI> top level DataObjects in a Datagraph are from the root table; therefore there is always a "root" table defined
 *   in relational metadata.</LI> 
 * 	<LI> AlbumCatalogJDBCSDOService uses this class to create metadata for queries. For example, for the findAlbum
 *   operation, AlbumCatalogJDBCSDOService calls the getAlbumMetadata method which has the Album table definition.</LI>
 * 	<LI> note that relationships between tables are defined as "Foreign Key" relationships in the metadata. This does NOT
 *   imply that the tables have a foreign key constraint in the actual database. A foreign key relationship in jdbc mediator
 *   relational metadata is used to specify a join between tables; it is not used to enforce referential integrity.</LI>
 *      
 * </UL> 
 *       	 
 */

public class AlbumCatalogMetadata {
	
	// names of tables and columns in the database
	public static final String BANDTABLE = "BAND";
	public static final String BAND_NAME_FIELD = "NAME";
	public static final String BAND_GENRE_FIELD = "GENRE";
	
	public static final String ALBUMTABLE = "ALBUM";
	public static final String ALBUM_BANDNAME_FIELD = "BANDNAME";
	public static final String ALBUM_NAME_FIELD = "ALBUMNAME";
	public static final String ALBUM_YEAR_FIELD = "ALBUMYEAR";
	public static final String ALBUM_TRACKS_FIELD = "RECOMMENDEDTRACKS";
	public static final String ALBUM_OCC_FIELD = "OCC_COUNT";
	
	
	// a name is given to a relationship between two tables; the following is the relationship name 
	//between the band and album table  
	public static final String BANDALBUM_REL = "BAND_ALBUM";
	
			
	/**
	 * Returns metadata for the Band table.  
	 * @return Metadata       	 
	 */
	public static Metadata getBandMetadata()
	{
		Metadata bandMetadata = factory().createMetadata();
		addBandTable(bandMetadata, true);
		
		return bandMetadata;
	}

	/**
	 * Returns metadata for the Band and Album table with a relationship between the two tables.
	 * Note that metadata for every table used in the query must be added.
	 *
 	 * @param  isBandRoot specifies if top level DataObjects in Datagraph are going to be from the Band table. 
	 * @param  isAlbumRelationshipExclusive an exclusive relationship does a left outer join between the two tables and an inclusive
	 *          relationship does an inner join. Therefore, when a query is done with this metadata, even if a band has no related albums
	 *         in the album table, it will still be included in the Datagraph.   
	 * @return Metadata       	 
	 */
	public static Metadata getBandAndAlbumMetadata(boolean isBandRoot, boolean isAlbumRelationshipExclusive)
	{
		
		Metadata bandWithAlbumMetadata = factory().createMetadata();
		
		if(isBandRoot)
		{
			addBandTable(bandWithAlbumMetadata, true);
			addAlbumTable(bandWithAlbumMetadata, false);
		}
		else 
		{
			addBandTable(bandWithAlbumMetadata, false);
			addAlbumTable(bandWithAlbumMetadata, true);
		} 
		
		addAlbumFKRelationship(bandWithAlbumMetadata, isAlbumRelationshipExclusive);
		
		return bandWithAlbumMetadata;		 		
	}
	
	/**
	 * Returns metadata for the Album table.
	 * 	    
	 * @return Metadata        	 
	 */
	public static Metadata getAlbumMetadata()
	{
		Metadata albumMetadata = factory().createMetadata();
		addAlbumTable(albumMetadata, true);
		
		return albumMetadata;
	}

	/**
	 * Adds Band table metadata, which includes columns and primary key definitions.
	 * 	    
	 * @param metadata the metadata to which the Band table definition is added
	 * @param isRoot specifies if top level DataObjects in Datagraph are going to be from the Band table.	         	
	 */
	public static void addBandTable(Metadata metadata, boolean isRoot)
	{
		
		// note that table and column names in metadata have to match their names in the database. 
		Table bandTable = metadata.addTable(BANDTABLE);
				 
		Column nameColumn = bandTable.addStringColumn(BAND_NAME_FIELD);		
		nameColumn.setNullable(false);
		
		bandTable.addStringColumn(BAND_GENRE_FIELD);
				
		//a column specified as the primary key in metadata, has to be the primary key in the actual database.      	
		bandTable.setPrimaryKey(nameColumn);
		
		if(isRoot)
			metadata.setRootTable(bandTable);
	}

	/**
	 * Adds Album table metadata
	 * 	    
	 * @param metadata the metadata to which the Album table definition is added
	 * @param isRoot specifies if top level DataObject in Datagraph are going to be from the Album table.	         	
	 */
	public static void addAlbumTable(Metadata metadata, boolean isRoot)
	{
		Table albumTable = metadata.addTable(ALBUMTABLE);
		
		Column bandName =  albumTable.addStringColumn(ALBUM_BANDNAME_FIELD);		
		bandName.setNullable(false);
		
		Column albumName = albumTable.addStringColumn(ALBUM_NAME_FIELD);
		albumName.setNullable(false);

		albumTable.addIntegerColumn(ALBUM_YEAR_FIELD);	
		albumTable.addStringColumn(ALBUM_TRACKS_FIELD);
				
		
		// Album table is the only table in the sample, that has a field which can be updated from the GUI.
		// Therefore, it is used to demonstrate the Optimistic Concurrency Control feature of JDBC Mediator. Look at technotes
		// of this sample in the Samples Gallery for details.  
		Column collisionColumn = albumTable.addIntegerColumn(ALBUM_OCC_FIELD);
		collisionColumn.setNullable(false);
		albumTable.setCollisionColumn(collisionColumn);
		
		//If the primary key in the database is comprised of more than one column, the Key metadata type can be used to
		//specify a multi-column key in the metadata. 
		Key albumTableKey = factory().createKey();
		albumTableKey.getColumns().add(bandName);
		albumTableKey.getColumns().add(albumName);
		albumTable.setPrimaryKey(albumTableKey);
		
		if(isRoot)
			metadata.setRootTable(albumTable);
	}
	
	/**
	 * Adds a relationship between the Band and Album table. A relationship is defined as a foreign key in the Album table.
	 * Note that a foreign key in JDBC Mediator relational metadata does NOT mean that there has to be a corresponding foreign key constraint in  
	 * the database; i.e. foreign keys in jdbc mediator relational metadata are not used for referential integrity, their purpose is
	 * to enable joins between fields of two tables, so related data can be browsed between DataObjects. 
	 *  	    
	 * @param metadata the metadata to which the relationship definition is added
	 * @param isExclusive specifies if an outer join or an inner join behavior is required. If true, an inner join will be performed; 
	 * 		 if false, a left outer join is performed 
	 *        	 
	 */
	public static void addAlbumFKRelationship(Metadata metadata, boolean isExclusive)
	{
		
		Table bandTable = metadata.getTable(BANDTABLE);		
		Table albumTable = metadata.getTable(ALBUMTABLE);
		
		// the column on which to do a join between the band and album table 
		Column bandRelationshipColumn = albumTable.getColumn(ALBUM_BANDNAME_FIELD);		
	
		
		// create a new key; this will be added as a foreign key in the album table 
		Key albumForeignKey = factory().createKey();
		
		//add the column on which to do a join, to the foreign key
		albumForeignKey.getColumns().add(bandRelationshipColumn);
		
		//add foreign key to album table
		albumTable.getForeignKeys().add(albumForeignKey);
		
		// a Relationship is defined between keys of two tables; in this case, the relationship is between
		// the primary key of Band (the NAME field in band table) and a foreign key of Album (the BANDNAME) field in album table  
		Key bandPK = bandTable.getPrimaryKey();						 
		Relationship band_albumRelationship = metadata.addRelationship(bandPK, albumForeignKey);
		
		//the exclusive property specifies an inner or outer join; if not specified, the default is true 
		band_albumRelationship.setExclusive(isExclusive);
		
		//a relationship can be given a name, this allows for browsing of related data in a DataObject. eg: bandDataObject.get(BANDALBUM_REL).
		band_albumRelationship.setName(BANDALBUM_REL);

	}

	/**
	 * Adds a filter on the NAME field of the Band table. A filter allows a parameter used in a SQL query to be
	 * defined at the time when a Datagraph is retrieved from a JDBC Mediator 
	 * 	    
	 * @param metadata the metadata to which the filter definition is added
	 */
	public static void addBandNameFilter(Metadata metadata)
	{
		Table bandTable = metadata.getTable(BANDTABLE);
		
		Filter filter = factory().createFilter();

		// The predicate string of a filter is exactly like a WHERE clause of a SQL statement. It is not parsed or interpreted by the
		// jdbc mediator, it is used as is. Only columns of tables defined in the metadata can be used in the predicate.
		filter.setPredicate(BAND_NAME_FIELD + " LIKE ?");

		// a filter argument describes the arguments for the predicate above. The getGraph method of the mediator, takes these
		// arguments in as a 
		// DataObject, where each property of the DataObject is the name of an argument. Look at findBand method in AlbumCatalogJDBCSDOService
		// for an example.                  
		FilterArgument arg = factory().createFilterArgument();
		arg.setName("bandName");
		arg.setType(Column.STRING);
		filter.getFilterArguments().add(arg);

		bandTable.setFilter(filter);
		
	}
		
	/**
	 * Returns the Metadata factory which is used to create Metadata artifacts.   
	 * 	    
	 */
	private static MetadataFactory factory()
	{
		return MetadataFactory.eINSTANCE;
	}
	

}
